#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp


# Setup the command line arguments.
parser = ArgumentParser()

# Output verbosity options.
parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                    action="store_const", dest="loglevel",
                    const=logging.ERROR, default=logging.INFO)
parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                    action="store_const", dest="loglevel",
                    const=logging.DEBUG, default=logging.INFO)

# JID and password options.
parser.add_argument("--oldjid", dest="old_jid",
                    help="JID of the old account")
parser.add_argument("--oldpassword", dest="old_password",
                    help="password of the old account")

parser.add_argument("--newjid", dest="new_jid",
                    help="JID of the old account")
parser.add_argument("--newpassword", dest="new_password",
                    help="password of the old account")


args = parser.parse_args()

# Setup logging.
logging.basicConfig(level=args.loglevel,
                    format='%(levelname)-8s %(message)s')

if args.old_jid is None:
    args.old_jid = input("Old JID: ")
if args.old_password is None:
    args.old_password = getpass("Old Password: ")

if args.new_jid is None:
    args.new_jid = input("New JID: ")
if args.new_password is None:
    args.new_password = getpass("New Password: ")


old_xmpp = slixmpp.ClientXMPP(args.old_jid, args.old_password)

# If you are connecting to Facebook and wish to use the
# X-FACEBOOK-PLATFORM authentication mechanism, you will need
# your API key and an access token. Then you'll set:
# xmpp.credentials['api_key'] = 'THE_API_KEY'
# xmpp.credentials['access_token'] = 'THE_ACCESS_TOKEN'

# If you are connecting to MSN, then you will need an
# access token, and it does not matter what JID you
# specify other than that the domain is 'messenger.live.com',
# so '_@messenger.live.com' will work. You can specify
# the access token as so:
# xmpp.credentials['access_token'] = 'THE_ACCESS_TOKEN'

# If you are working with an OpenFire server, you may need
# to adjust the SSL version used:
# xmpp.ssl_version = ssl.PROTOCOL_SSLv3

# If you want to verify the SSL certificates offered by a server:
# xmpp.ca_certs = "path/to/ca/cert"

roster = []

async def on_session(event):
    roster.append(await old_xmpp.get_roster())
    old_xmpp.disconnect()
old_xmpp.add_event_handler('session_start', on_session)

if old_xmpp.connect():
    asyncio.get_event_loop().run_until_complete(old_xmpp.disconnected)

if not roster:
    print('No roster to migrate')
    sys.exit()

new_xmpp = slixmpp.ClientXMPP(args.new_jid, args.new_password)
async def on_session2(event):
    await new_xmpp.get_roster()
    new_xmpp.send_presence()

    logging.info(roster[0])
    data = roster[0]['roster']['items']
    logging.info(data)

    for jid, item in data.items():
        if item['subscription'] != 'none':
            new_xmpp.send_presence(ptype='subscribe', pto=jid)
        await new_xmpp.update_roster(
            jid,
            name=item['name'],
            groups=item['groups']
        )
    new_xmpp.disconnect()
new_xmpp.add_event_handler('session_start', on_session2)

new_xmpp.connect()
asyncio.get_event_loop().run_until_complete(new_xmpp.disconnected)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from __future__ import annotations

import hashlib
import logging

from asyncio import Future
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Union,
    TYPE_CHECKING,
)

from slixmpp.stanza import Iq
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin, JID
from slixmpp.plugins.xep_0084.stanza import Data, MetaData, Pointer
from slixmpp.plugins.xep_0084 import stanza

try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict


class AvatarMetadataItem(TypedDict, total=False):
    bytes: int
    id: str
    type: str
    height: int
    width: int
    url: str

MetadataItems = Union[
    AvatarMetadataItem,
    List[AvatarMetadataItem],
    Set[AvatarMetadataItem]
]


log = logging.getLogger(__name__)


class XEP_0084(BasePlugin):

    name = 'xep_0084'
    description = 'XEP-0084: User Avatar'
    dependencies = {'xep_0163', 'xep_0060'}
    stanza = stanza

    def plugin_init(self):
        pubsub_stanza = self.xmpp['xep_0060'].stanza
        register_stanza_plugin(pubsub_stanza.Item, Data)
        register_stanza_plugin(pubsub_stanza.EventItem, Data)

        self.xmpp['xep_0060'].map_node_event(Data.namespace, 'avatar_data')

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=MetaData.namespace)
        self.xmpp['xep_0163'].remove_interest(MetaData.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('avatar_metadata', MetaData)

    def generate_id(self, data) -> str:
        return hashlib.sha1(data).hexdigest()

    def retrieve_avatar(self, jid: JID, id: str, **pubsubkwargs) -> Future:
        """Retrieve an avatar.

        :param jid: JID of the entity to get the avatar from.
        :param id: Identifier of the item containing the avatar.
        """
        return self.xmpp['xep_0060'].get_item(
            jid,
            Data.namespace,
            id,
            **pubsubkwargs
        )

    def publish_avatar(self, data: bytes, **pubsubkwargs) -> Future:
        """Publish an avatar.

        :param data: The avatar, in bytes representation.
        """
        payload = Data()
        payload['value'] = data
        return self.xmpp['xep_0163'].publish(
            payload,
            id=self.generate_id(data),
            **pubsubkwargs
        )

    def publish_avatar_metadata(self, items: Optional[MetadataItems] = None,
                                pointers: Optional[Iterable[Pointer]] = None,
                                **pubsubkwargs) -> Future:
        """Publish avatar metadata.

        :param items: Metadata items to store
        :param pointers: Optional pointers
        """
        metadata = MetaData()
        if items is None:
            items = []
        if not isinstance(items, (list, set)):
            items = [items]
        for info in items:
            metadata.add_info(info['id'], info['type'], info['bytes'],
                    height=info.get('height', ''),
                    width=info.get('width', ''),
                    url=info.get('url', ''))

        if pointers is not None:
            for pointer in pointers:
                metadata.add_pointer(pointer)

        return self.xmpp['xep_0163'].publish(
            metadata,
            id=info['id'],
            **pubsubkwargs
        )

    def stop(self, **pubsubkwargs) -> Future:
        """
        Clear existing avatar metadata information to stop notifications.
        """
        metadata = MetaData()
        return self.xmpp['xep_0163'].publish(
            metadata,
            node=MetaData.namespace,
            **pubsubkwargs
        )



from datetime import datetime
from typing import Optional

from slixmpp.plugins.xep_0082 import format_datetime, parse
from slixmpp.plugins.xep_0300 import Hash
from slixmpp.plugins.xep_0264.stanza import Thumbnail
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = "urn:xmpp:file:metadata:0"


class File(ElementBase):
    name = "file"
    namespace = NS
    plugin_attrib = "file"
    interfaces = sub_interfaces = {
        "media-type",
        "name",
        "date",
        "size",
        "desc",
        "width",
        "height",
        "length"
    }

    def set_width(self, width: int):
        self.__set_if_positive("width", width)

    def get_width(self) -> Optional[int]:
        return _positive_int_or_none(self._get_sub_text("width"))

    def set_height(self, height: int):
        self.__set_if_positive("height", height)

    def get_height(self) -> Optional[int]:
        return _positive_int_or_none(self._get_sub_text("height"))

    def set_length(self, length: int):
        self.__set_if_positive("length", length)

    def get_length(self) -> Optional[int]:
        return _positive_int_or_none(self._get_sub_text("length"))

    def set_size(self, size: int):
        self.__set_if_positive("size", size)

    def get_size(self) -> Optional[int]:
        return _positive_int_or_none(self._get_sub_text("size"))

    def get_date(self) -> Optional[datetime]:
        try:
            return parse(self._get_sub_text("date"))
        except ValueError:
            return

    def set_date(self, stamp: datetime):
        try:
            self._set_sub_text("date", format_datetime(stamp))
        except ValueError:
            pass

    def __set_if_positive(self, key: str, value: int):
        if value <= 0:
            raise ValueError(f"Invalid value for element {key}: {value}")
        self._set_sub_text(key, str(value))


def _positive_int_or_none(v: str) -> Optional[int]:
    try:
        return int(v)
    except ValueError:
        return None

def register_plugins():
    register_stanza_plugin(File, Hash)
    register_stanza_plugin(File, Thumbnail)



import unittest

from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0469 import stanza
from slixmpp.plugins.xep_0402 import stanza as b_stanza


class TestBookmarksPinning(SlixTest):
    def setUp(self):
        b_stanza.register_plugin()
        stanza.register_plugin()

    def test_pinned(self):
        bookmark = b_stanza.Conference()
        bookmark["password"] = "pass"
        bookmark["nick"] = "nick"
        bookmark["autojoin"] = False
        bookmark["extensions"].enable("pinned")
        self.check(
            bookmark,
            """
            <conference xmlns='urn:xmpp:bookmarks:1'
                        autojoin='false'>
              <nick>nick</nick>
              <password>pass</password>
              <extensions>
                <pinned xmlns="urn:xmpp:bookmarks-pinning:0" />
              </extensions>
            </conference>
            """,
            use_values=False
        )



suite = unittest.TestLoader().loadTestsFromTestCase(TestBookmarksPinning)



import unittest
from slixmpp import JID, Iq, Message
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0313 import stanza
from slixmpp.plugins.xep_0004.stanza import Form
from slixmpp.plugins.xep_0297 import stanza as fstanza
from slixmpp.plugins.xep_0059 import stanza as rstanza
from slixmpp.xmlstream import register_stanza_plugin


class TestMAM(SlixTest):

    def setUp(self):
        register_stanza_plugin(stanza.MAM, Form)
        register_stanza_plugin(Iq, stanza.MAM)
        register_stanza_plugin(Message, stanza.Result)
        register_stanza_plugin(Iq, stanza.Fin)
        register_stanza_plugin(
            stanza.Result,
            fstanza.Forwarded
        )
        register_stanza_plugin(stanza.MAM, rstanza.Set)
        register_stanza_plugin(stanza.Fin, rstanza.Set)

        register_stanza_plugin(Iq, stanza.Metadata)
        register_stanza_plugin(stanza.Metadata, stanza.Start)
        register_stanza_plugin(stanza.Metadata, stanza.End)

    def testMAMQuery(self):
        """Test that we can build a simple MAM query."""
        iq = Iq()
        iq['type'] = 'set'
        iq['mam']['queryid'] = 'f27'

        self.check(iq, """
            <iq type='set'>
              <query xmlns='urn:xmpp:mam:2' queryid='f27'/>
            </iq>
        """)

    def testMAMQueryOptions(self):
        """Test that we can build a mam query with all options."""
        iq = Iq()
        iq['type'] = 'set'
        iq['mam']['with'] = JID('juliet@capulet.lit')
        iq['mam']['start'] = '2010-06-07T00:00:00Z'
        iq['mam']['end'] = '2010-07-07T13:23:54Z'
        iq['mam']['after_id'] = 'id1'
        iq['mam']['before_id'] = 'id2'
        iq['mam']['ids'] = ['a', 'b', 'c']

        self.check(iq, """
            <iq type='set'>
              <query xmlns='urn:xmpp:mam:2'>
                <x xmlns='jabber:x:data' type='submit'>
                  <field var='FORM_TYPE' type='hidden'>
                    <value>urn:xmpp:mam:2</value>
                  </field>
                  <field var='with'>
                    <value>juliet@capulet.lit</value>
                  </field>
                  <field var='start'>
                    <value>2010-06-07T00:00:00Z</value>
                  </field>
                  <field var='end'>
                    <value>2010-07-07T13:23:54Z</value>
                  </field>
                  <field var='after-id'>
                    <value>id1</value>
                  </field>
                  <field var='before-id'>
                    <value>id2</value>
                  </field>
                  <field var='ids'>
                    <value>a</value>
                    <value>b</value>
                    <value>c</value>
                  </field>
                </x>
              </query>
            </iq>
        """, use_values=False)

    def testMAMMetadata(self):
        """Test that we can build a MAM metadata payload"""

        iq = Iq()
        iq['type'] = 'result'
        iq['mam_metadata']['start']['id'] = 'YWxwaGEg'
        iq['mam_metadata']['start']['timestamp'] = '2008-08-22T21:09:04Z'
        iq['mam_metadata']['end']['id'] = 'b21lZ2Eg'
        iq['mam_metadata']['end']['timestamp'] = '2020-04-20T14:34:21Z'

        self.check(iq, """
            <iq type='result'>
              <metadata xmlns='urn:xmpp:mam:2'>
                <start id='YWxwaGEg' timestamp='2008-08-22T21:09:04Z' />
                <end id='b21lZ2Eg' timestamp='2020-04-20T14:34:21Z' />
              </metadata>
            </iq>
        """)



suite = unittest.TestLoader().loadTestsFromTestCase(TestMAM)



import unittest
from datetime import datetime
from base64 import b64encode
from pathlib import Path
from tempfile import NamedTemporaryFile
from hashlib import sha256

from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0082 import format_datetime


class TestStatelessFileSharing(SlixTest):
    def setUp(self):
        self.stream_start(
            mode="component", jid="whatevs.shakespeare.lit", plugins={"xep_0447"}
        )


    def test_set_file(self):
        with NamedTemporaryFile("wb+") as f:
            n = 10
            size = 0
            for i in range(n):
                size += len(bytes(i))
                f.write(bytes(i))

            f.seek(0)
            h = b64encode(sha256(f.read()).digest()).decode()
            sfs = self.xmpp["xep_0447"].get_sfs(
                Path(f.name),
                ["https://xxx.com"],
                media_type="MEDIA",
                desc="DESCRIPTION",
            )

            self.check(
                sfs,
                f"""
                  <file-sharing xmlns='urn:xmpp:sfs:0' disposition='inline'>
                    <file xmlns='urn:xmpp:file:metadata:0'>
                      <media-type>MEDIA</media-type>
                      <name>{Path(f.name).name}</name>
                      <size>{size}</size>
                      <hash xmlns='urn:xmpp:hashes:2' algo='sha-256'>{h}</hash>
                      <desc>DESCRIPTION</desc>
                      <date>{format_datetime(datetime.fromtimestamp(Path(f.name).stat().st_mtime))}</date>
                    </file>
                    <sources>
                      <url-data xmlns='http://jabber.org/protocol/url-data'
                                target='https://xxx.com' />
                    </sources>
                  </file-sharing>
                """,
                use_values=False,
            )

suite = unittest.TestLoader().loadTestsFromTestCase(TestStatelessFileSharing)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
import logging

import slixmpp
from slixmpp.stanza import Message
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin, ElementBase, ET
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0085 import stanza, ChatState


log = logging.getLogger(__name__)


class XEP_0085(BasePlugin):

    """
    XEP-0085 Chat State Notifications
    """

    name = 'xep_0085'
    description = 'XEP-0085: Chat State Notifications'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_handler(
            Callback('Chat State',
                     StanzaPath('message/chat_state'),
                     self._handle_chat_state))

        register_stanza_plugin(Message, stanza.Active)
        register_stanza_plugin(Message, stanza.Composing)
        register_stanza_plugin(Message, stanza.Gone)
        register_stanza_plugin(Message, stanza.Inactive)
        register_stanza_plugin(Message, stanza.Paused)

    def plugin_end(self):
        self.xmpp.remove_handler('Chat State')

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(ChatState.namespace)

    def _handle_chat_state(self, msg):
        state = msg['chat_state']
        log.debug("Chat State: %s, %s", state, msg['from'].jid)
        self.xmpp.event('chatstate', msg)
        self.xmpp.event('chatstate_%s' % state, msg)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.xmlstream import ElementBase, ET


class UserMood(ElementBase):

    name = 'mood'
    namespace = 'http://jabber.org/protocol/mood'
    plugin_attrib = 'mood'
    interfaces = {'value', 'text'}
    sub_interfaces = {'text'}
    moods = {'afraid', 'amazed', 'amorous', 'angry', 'annoyed', 'anxious',
             'aroused', 'ashamed', 'bored', 'brave', 'calm', 'cautious',
             'cold', 'confident', 'confused', 'contemplative', 'contented',
             'cranky', 'crazy', 'creative', 'curious', 'dejected',
             'depressed', 'disappointed', 'disgusted', 'dismayed',
             'distracted', 'embarrassed', 'envious', 'excited',
             'flirtatious', 'frustrated', 'grateful', 'grieving', 'grumpy',
             'guilty', 'happy', 'hopeful', 'hot', 'humbled', 'humiliated',
             'hungry', 'hurt', 'impressed', 'in_awe', 'in_love',
             'indignant', 'interested', 'intoxicated', 'invincible',
             'jealous', 'lonely', 'lost', 'lucky', 'mean', 'moody',
             'nervous', 'neutral', 'offended', 'outraged', 'playful',
             'proud', 'relaxed', 'relieved', 'remorseful', 'restless',
             'sad', 'sarcastic', 'satisfied', 'serious', 'shocked',
             'shy', 'sick', 'sleepy', 'spontaneous', 'stressed', 'strong',
             'surprised', 'thankful', 'thirsty', 'tired', 'undefined',
             'weak', 'worried'}

    def set_value(self, value):
        self.del_value()
        if value in self.moods:
            self._set_sub_text(value, '', keep=True)
        else:
            raise ValueError('Unknown mood value')

    def get_value(self):
        for child in self.xml:
            if child.tag.startswith('{%s}' % self.namespace):
                elem_name = child.tag.split('}')[-1]
                if elem_name in self.moods:
                    return elem_name
        return ''

    def del_value(self):
        curr_value = self.get_value()
        if curr_value:
            self._set_sub_text(curr_value, '', keep=False)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.stanza import Message
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)


NS = 'urn:xmpp:tmp:quick-response'


class Response(ElementBase):
    namespace = NS
    name = 'response'
    plugin_attrib = 'response'
    interfaces = {'value', 'label'}


class Action(ElementBase):
    namespace = NS
    name = 'action'
    plugin_attrib = 'action'
    interfaces = {'id', 'label'}


class ActionSelected(ElementBase):
    namespace = NS
    name = 'action-selected'
    plugin_attrib = 'action_selected'
    interfaces = {'id'}


def register_plugins():
    register_stanza_plugin(Message, Action, iterable=True)
    register_stanza_plugin(Message, ActionSelected)
    register_stanza_plugin(Message, Response, iterable=True)




# slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase

NS ='urn:xmpp:chat-markers:0'

class Markable(ElementBase):
    name = 'markable'
    plugin_attrib = 'markable'
    namespace = NS


class Received(ElementBase):
    name = 'received'
    plugin_attrib = 'received'
    namespace = NS
    interfaces = {'id'}


class Displayed(ElementBase):
    name = 'displayed'
    plugin_attrib = 'displayed'
    namespace = NS
    interfaces = {'id'}


class Acknowledged(ElementBase):
    name = 'acknowledged'
    plugin_attrib = 'acknowledged'
    namespace = NS
    interfaces = {'id'}



import unittest
from slixmpp.test import SlixTest

from slixmpp.plugins.xep_0356 import stanza, permissions


class TestPermissions(SlixTest):
    def setUp(self):
        stanza.register()

    def testAdvertisePermission(self):
        xmlstring = """
            <message from='capulet.lit' to='pubsub.capulet.lit'>
                <privilege xmlns='urn:xmpp:privilege:2'>
                    <perm access='roster' type='both'/>
                    <perm access='message' type='outgoing'/>
                    <perm access='presence' type='managed_entity'/>
                    <perm access='iq' type='both'/>
                </privilege>
            </message>
        """
        msg = self.Message()
        msg["from"] = "capulet.lit"
        msg["to"] = "pubsub.capulet.lit"

        for access, type_ in [
            ("roster", permissions.RosterAccess.BOTH),
            ("message", permissions.MessagePermission.OUTGOING),
            ("presence", permissions.PresencePermission.MANAGED_ENTITY),
            ("iq", permissions.IqPermission.BOTH),
        ]:
            msg["privilege"].add_perm(access, type_)

        self.check(msg, xmlstring)

    def testIqPermission(self):
        x = stanza.Privilege()
        x["access"] = "iq"
        ns = stanza.NameSpace()
        ns["ns"] = "some_ns"
        ns["type"] = "get"
        x["perm"]["access"] = "iq"
        x["perm"].append(ns)
        ns = stanza.NameSpace()
        ns["ns"] = "some_other_ns"
        ns["type"] = "both"
        x["perm"].append(ns)
        self.check(
            x,
            """
              <privilege xmlns='urn:xmpp:privilege:2'>
                <perm access='iq'>
                  <namespace ns='some_ns' type='get' />
                  <namespace ns='some_other_ns' type='both' />
                </perm>
              </privilege>
            """
        )
        nss = set()
        for perm in x["perms"]:
            for ns in perm["namespaces"]:
                nss.add((ns["ns"], ns["type"]))
        assert nss == {("some_ns", "get"), ("some_other_ns", "both")}


suite = unittest.TestLoader().loadTestsFromTestCase(TestPermissions)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
from os import getenv

# Use defusedxml if wanted
# Since enabling it can have adverse consequences for the programs using
# slixmpp, do not enable it by default.
if  getenv('SLIXMPP_ENABLE_DEFUSEDXML', default='false').lower() == 'true':
    try:
        import defusedxml
        defusedxml.defuse_stdlib()
    except ImportError:
        pass

from slixmpp.stanza import Message, Presence, Iq
from slixmpp.jid import JID, InvalidJID
from slixmpp.xmlstream.stanzabase import ET, ElementBase, register_stanza_plugin
from slixmpp.xmlstream.handler import *
from slixmpp.xmlstream import XMLStream
from slixmpp.xmlstream.matcher import *
from slixmpp.basexmpp import BaseXMPP
from slixmpp.clientxmpp import ClientXMPP
from slixmpp.componentxmpp import ComponentXMPP

from slixmpp.version import __version__, __version_info__

__all__ = [
    'Message', 'Presence', 'Iq', 'JID', 'InvalidJID', 'ET', 'ElementBase',
    'register_stanza_plugin', 'XMLStream', 'BaseXMPP', 'ClientXMPP', 'ComponentXMPP',
    '__version__', '__version_info__'
]




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin, BasePlugin

from slixmpp.plugins.google.gmail import Gmail
from slixmpp.plugins.google.auth import GoogleAuth
from slixmpp.plugins.google.settings import GoogleSettings
from slixmpp.plugins.google.nosave import GoogleNoSave


class Google(BasePlugin):

    """
    Google: Custom GTalk Features

    Also see: <https://developers.google.com/talk/jep_extensions/extensions>
    """

    name = 'google'
    description = 'Google: Custom GTalk Features'
    dependencies = set([
        'gmail',
        'google_settings',
        'google_nosave',
        'google_auth'
    ])

    def __getitem__(self, attr):
        if attr in ('settings', 'nosave', 'auth'):
            return self.xmpp['google_%s' % attr]
        elif attr == 'gmail':
            return self.xmpp['gmail']
        else:
            raise KeyError(attr)


register_plugin(Gmail)
register_plugin(GoogleAuth)
register_plugin(GoogleSettings)
register_plugin(GoogleNoSave)
register_plugin(Google)



import time
import unittest
from slixmpp.test import SlixTest


class TestStreamTester(SlixTest):
    """
    Test that we can simulate and test a stanza stream.
    """

    def testClientEcho(self):
        """Test that we can interact with a ClientXMPP instance."""
        self.stream_start(mode='client')

        def echo(msg):
            msg.reply('Thanks for sending: %s' % msg['body']).send()

        self.xmpp.add_event_handler('message', echo)

        self.recv("""
          <message to="tester@localhost" from="user@localhost">
            <body>Hi!</body>
          </message>
        """)

        self.send("""
          <message to="user@localhost">
            <body>Thanks for sending: Hi!</body>
          </message>
        """)

    def testComponentEcho(self):
        """Test that we can interact with a ComponentXMPP instance."""
        self.stream_start(mode='component')

        def echo(msg):
            msg.reply('Thanks for sending: %(body)s' % msg).send()

        self.xmpp.add_event_handler('message', echo)

        self.recv("""
          <message to="tester.localhost" from="user@localhost">
            <body>Hi!</body>
          </message>
        """)

        self.send("""
          <message to="user@localhost" from="tester.localhost">
            <body>Thanks for sending: Hi!</body>
          </message>
        """)

    def testSendStreamHeader(self):
        """Test that we can check a sent stream header."""
        self.stream_start(mode='client', skip=False)
        self.send_header(sto='localhost')

suite = unittest.TestLoader().loadTestsFromTestCase(TestStreamTester)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class Reachability(ElementBase):
    name = 'reach'
    namespace = 'urn:xmpp:reach:0'
    plugin_attrib = 'reach'
    interfaces = set()


class Address(ElementBase):
    name = 'addr'
    namespace = 'urn:xmpp:reach:0'
    plugin_attrib = 'address'
    plugin_multi_attrib = 'addresses'
    interfaces = {'uri', 'desc'}
    lang_interfaces = {'desc'}
    sub_interfaces = {'desc'}


register_stanza_plugin(Reachability, Address, iterable=True)



import unittest
from slixmpp import Message
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0439 import stanza


class TestQuickResponse(SlixTest):

    def setUp(self):
        stanza.register_plugins()

    def testResponse(self):
        message = Message()
        message['body'] = 'Reply 1 or 2?'
        for (value, label) in [('1', 'Rep 1'), ('2', 'Rep 2')]:
            rep = stanza.Response()
            rep['value'] = value
            rep['label'] = label
            message.append(rep)

        self.check(message, """
<message>
  <body>Reply 1 or 2?</body>
  <response xmlns="urn:xmpp:tmp:quick-response" value="1" label="Rep 1" />
  <response xmlns="urn:xmpp:tmp:quick-response" value="2" label="Rep 2" />
</message>
        """, use_values=False)

    def testAction(self):
        message = Message()
        message['body'] = 'action 1 or 2?'
        for (id_, label) in [('1', 'action 1'), ('2', 'action 2')]:
            act = stanza.Action()
            act['id'] = id_
            act['label'] = label
            message.append(act)

        self.check(message, """
<message>
  <body>action 1 or 2?</body>
  <action xmlns="urn:xmpp:tmp:quick-response" id="1" label="action 1" />
  <action xmlns="urn:xmpp:tmp:quick-response" id="2" label="action 2" />
</message>
        """, use_values=False)

    def testActionSelected(self):
        message = Message()
        message['action_selected']['id'] = 'act1'

        self.check(message, """
<message>
  <action-selected xmlns="urn:xmpp:tmp:quick-response" id="act1" />
</message>
        """, use_values=False)


suite = unittest.TestLoader().loadTestsFromTestCase(TestQuickResponse)



import unittest

from slixmpp import register_stanza_plugin, Iq
from slixmpp.test import SlixTest

from slixmpp.plugins.xep_0055 import stanza


class TestJabberSearch(SlixTest):
    def setUp(self):
        register_stanza_plugin(Iq, stanza.Search)
        self.stream_start(plugins={"xep_0055"})

    def testRequestSearchFields(self):
        iq = self.Iq()
        iq.set_from("juliet@capulet.com/balcony")
        iq.set_to("characters.shakespeare.lit")
        iq.set_type("get")
        iq.enable("search")
        iq["id"] = "0"
        self.check(
            iq,
            """
            <iq type='get'
                from='juliet@capulet.com/balcony'
                to='characters.shakespeare.lit'>
              <query xmlns='jabber:iq:search'/>
            </iq>
            """,
        )

    def testSendSearch(self):
        iq = self.xmpp["xep_0055"].make_search_iq(
            ifrom="juliet@capulet.com/balcony", ito="characters.shakespeare.lit"
        )
        iq["search"]["form"].add_field(var="x-gender", value="male")
        iq["id"] = "0"
        self.check(
            iq,
            """
            <iq type='set'
                from='juliet@capulet.com/balcony'
                to='characters.shakespeare.lit'>
              <query xmlns='jabber:iq:search'>
                <x xmlns='jabber:x:data' type='submit'>
                  <field type='hidden' var='FORM_TYPE'>
                    <value>jabber:iq:search</value>
                  </field>
                  <field var='x-gender'>
                    <value>male</value>
                  </field>
                </x>
              </query>
            </iq>
            """,
            use_values=False,
        )


suite = unittest.TestLoader().loadTestsFromTestCase(TestJabberSearch)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Presence
from slixmpp.xmlstream import JID
from slixmpp.roster import RosterNode
from slixmpp.types import RosterDBProtocol, JidStr

from typing import Iterable, Dict, Optional, TYPE_CHECKING, Iterator

if TYPE_CHECKING:
    from slixmpp import BaseXMPP


class Roster:

    """
    Slixmpp's roster manager.

    The roster is divided into "nodes", where each node is responsible
    for a single JID. While the distinction is not strictly necessary
    for client connections, it is a necessity for components that use
    multiple JIDs.

    Rosters may be stored and persisted in an external datastore. An
    interface object to the datastore that loads and saves roster items may
    be provided. See the documentation for the RosterItem class for the
    methods that the datastore interface object must provide.

    Attributes:
        xmpp           -- The main Slixmpp instance.
        db             -- Optional interface object to an external datastore.
        auto_authorize -- Default auto_authorize value for new roster nodes.
                          Defaults to True.
        auto_subscribe -- Default auto_subscribe value for new roster nodes.
                          Defaults to True.

    Methods:
        add           -- Create a new roster node for a JID.
        send_presence -- Shortcut for sending a presence stanza.
    """

    xmpp: "BaseXMPP"
    db: Optional[RosterDBProtocol]
    _auto_authorize: bool
    _auto_subscribe: bool
    _rosters: Dict[str, RosterNode]

    def __init__(self, xmpp: 'BaseXMPP', db: Optional[RosterDBProtocol] = None) -> None:
        """
        Create a new roster.

        :param xmpp: The main Slixmpp instance.
        :param db: Optional interface object to a datastore.
        """
        self.xmpp = xmpp
        self.db = db
        self._auto_authorize = True
        self._auto_subscribe = True
        self._rosters = {}

        if self.db:
            for node in self.db.entries(None, {}):
                self.add(node)

        self.xmpp.add_filter('out', self._save_last_status)

    def _save_last_status(self, stanza) -> None:

        if isinstance(stanza, Presence):
            sfrom = stanza['from'].full
            sto = stanza['to'].full

            if not sfrom:
                sfrom = self.xmpp.boundjid

            if stanza['type'] in stanza.showtypes or \
               stanza['type'] in ('available', 'unavailable'):
                if sto:
                    self[sfrom][sto].last_status = stanza
                else:
                    self[sfrom].last_status = stanza
                    for jid in self[sfrom]:
                        self[sfrom][jid].last_status = None

                if not self.xmpp.sentpresence:
                    self.xmpp.event('sent_presence')
                    self.xmpp.sentpresence = True

        return stanza

    def __getitem__(self, key: JidStr) -> RosterNode:
        """
        Return the roster node for a JID.

        A new roster node will be created if one
        does not already exist.

        :param key: Return the roster for this JID.
        """
        if key is None:
            key = self.xmpp.boundjid
        if not isinstance(key, JID):
            key = JID(key)
        key = key.bare

        if key not in self._rosters:
            self.add(key)
            self._rosters[key].auto_authorize = self.auto_authorize
            self._rosters[key].auto_subscribe = self.auto_subscribe
        return self._rosters[key]

    def keys(self) -> Iterable[str]:
        """Return the JIDs managed by the roster."""
        return self._rosters.keys()

    def __iter__(self) -> Iterator[str]:
        """Iterate over the roster nodes."""
        return self._rosters.__iter__()

    def add(self, node: JidStr) -> None:
        """
        Add a new roster node for the given JID.

        :param node: The JID for the new roster node.
        """
        if not isinstance(node, JID):
            node = JID(node)

        node = node.bare
        if node not in self._rosters:
            self._rosters[node] = RosterNode(self.xmpp, node, self.db)

    def set_backend(self, db: Optional[RosterDBProtocol] = None, save: bool = True) -> None:
        """
        Set the datastore interface object for the roster.

        :param db: The new datastore interface.
        :param save: If True, save the existing state to the new
                     backend datastore. Defaults to True.
        """
        self.db = db
        if self.db:
            existing_entries = set(self._rosters)
            new_entries = set(self.db.entries(None, {}))

            for node in existing_entries:
                self._rosters[node].set_backend(db, save)
            for node in new_entries - existing_entries:
                self.add(node)

    def reset(self) -> None:
        """
        Reset the state of the roster to forget any current
        presence information. Useful after a disconnection occurs.
        """
        for node in self:
            self[node].reset()

    def send_presence(self, **kwargs) -> None:
        """
        Create, initialize, and send a Presence stanza.

        If no recipient is specified, send the presence immediately.
        Otherwise, forward the send request to the recipient's roster
        entry for processing.

        :param pshow: The presence's show value.
        :param pstatus: The presence's status message.
        :param ppriority: This connections' priority.
        :param pto: The recipient of a directed presence.
        :param pfrom: The sender of a directed presence, which should
                      be the owner JID plus resource.
        :param ptype: The type of presence, such as 'subscribe'.
        :param pnick: Optional nickname of the presence's sender.
        """
        if self.xmpp.is_component and not kwargs.get('pfrom', ''):
            raise ValueError("Components must set the pfrom attribute!")
        self.xmpp.send_presence(**kwargs)

    @property
    def auto_authorize(self) -> bool:
        """
        Auto accept or deny subscription requests.

        If True, auto accept subscription requests.
        If False, auto deny subscription requests.
        If None, don't automatically respond.
        """
        return self._auto_authorize

    @auto_authorize.setter
    def auto_authorize(self, value: bool) -> None:
        """
        Auto accept or deny subscription requests.

        If True, auto accept subscription requests.
        If False, auto deny subscription requests.
        If None, don't automatically respond.
        """
        self._auto_authorize = value
        for node in self._rosters:
            self._rosters[node].auto_authorize = value

    @property
    def auto_subscribe(self) -> bool:
        """
        Auto send requests for mutual subscriptions.

        If True, auto send mutual subscription requests.
        """
        return self._auto_subscribe

    @auto_subscribe.setter
    def auto_subscribe(self, value: bool) -> None:
        """
        Auto send requests for mutual subscriptions.

        If True, auto send mutual subscription requests.
        """
        self._auto_subscribe = value
        for node in self._rosters:
            self._rosters[node].auto_subscribe = value

    def __repr__(self) -> str:
        return repr(self._rosters)



from slixmpp.xmlstream import ElementBase

NAMESPACE = "urn:xmpp:sims:1"


class Sims(ElementBase):
    name = "media-sharing"
    plugin_attrib = "sims"
    namespace = NAMESPACE


class Sources(ElementBase):
    name = plugin_attrib = "sources"
    namespace = NAMESPACE



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

__version__ = '1.9.1'
__version_info__ = (1, 9, 1)



import datetime
from typing import Optional

from slixmpp import ElementBase, Iq, register_stanza_plugin

NS = "urn:ietf:params:xml:ns:vcard-4.0"


class _VCardElementBase(ElementBase):
    namespace = NS


class VCard4(_VCardElementBase):
    name = plugin_attrib = "vcard"
    interfaces = {"full_name", "given", "surname", "birthday"}

    def set_full_name(self, full_name: str):
        self["fn"]["text"] = full_name

    def get_full_name(self):
        return self["fn"]["text"]

    def set_given(self, given: str):
        self["n"]["given"] = given

    def get_given(self):
        return self["n"]["given"]

    def set_surname(self, surname: str):
        self["n"]["surname"] = surname

    def get_surname(self):
        return self["n"]["surname"]

    def set_birthday(self, birthday: datetime.date):
        self["bday"]["date"] = birthday

    def get_birthday(self):
        return self["bday"]["date"]

    def add_tel(self, number: str, name: Optional[str] = None):
        tel = Tel()
        if name:
            tel["parameters"]["type_"]["text"] = name
        tel["uri"] = f"tel:{number}"
        self.append(tel)

    def add_address(
        self, country: Optional[str] = None, locality: Optional[str] = None
    ):
        adr = Adr()
        if locality:
            adr["locality"] = locality
        if country:
            adr["country"] = country
        self.append(adr)

    def add_nickname(self, nick: str):
        el = Nickname()
        el["text"] = nick
        self.append(el)

    def add_note(self, note: str):
        el = Note()
        el["text"] = note
        self.append(el)

    def add_impp(self, impp: str):
        el = Impp()
        el["uri"] = impp
        self.append(el)

    def add_url(self, url: str):
        el = Url()
        el["uri"] = url
        self.append(el)

    def add_email(self, email: str):
        el = Email()
        el["text"] = email
        self.append(el)


class _VCardTextElementBase(_VCardElementBase):
    interfaces = {"text"}
    sub_interfaces = {"text"}


class Fn(_VCardTextElementBase):
    name = plugin_attrib = "fn"


class Nickname(_VCardTextElementBase):
    name = plugin_attrib = "nickname"


class Note(_VCardTextElementBase):
    name = plugin_attrib = "note"


class _VCardUriElementBase(_VCardElementBase):
    interfaces = {"uri"}
    sub_interfaces = {"uri"}


class Url(_VCardUriElementBase):
    name = plugin_attrib = "url"


class Impp(_VCardUriElementBase):
    name = plugin_attrib = "impp"


class Email(_VCardTextElementBase):
    name = plugin_attrib = "email"


class N(_VCardElementBase):
    name = "n"
    plugin_attrib = "n"
    interfaces = sub_interfaces = {"given", "surname", "additional"}


class BDay(_VCardElementBase):
    name = plugin_attrib = "bday"
    interfaces = {"date"}

    def set_date(self, date: datetime.date):
        d = Date()
        d.xml.text = date.strftime("%Y-%m-%d")
        self.append(d)

    def get_date(self):
        for elem in self.xml:
            try:
                return datetime.date.fromisoformat(elem.text)
            except ValueError:
                return None


class Date(_VCardElementBase):
    name = "date"


class Tel(_VCardUriElementBase):
    name = plugin_attrib = "tel"


class Parameters(_VCardElementBase):
    name = plugin_attrib = "parameters"


class Type(_VCardTextElementBase):
    name = "type"
    plugin_attrib = "type_"


class Adr(_VCardElementBase):
    name = plugin_attrib = "adr"
    interfaces = sub_interfaces = {"locality", "country"}


register_stanza_plugin(Parameters, Type)
register_stanza_plugin(Tel, Parameters)
for p in N, Fn, Nickname, Note, Url, Impp, Email, BDay, Tel, Adr:
    register_stanza_plugin(VCard4, p, iterable=True)
register_stanza_plugin(Iq, VCard4)



import logging
import typing
import uuid
from collections import defaultdict

from slixmpp import JID, Iq, Message
from slixmpp.plugins.base import BasePlugin
from slixmpp.xmlstream import StanzaBase
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath

from . import stanza
from .permissions import IqPermission, MessagePermission, Permissions, RosterAccess

log = logging.getLogger(__name__)


class XEP_0356(BasePlugin):
    """
    XEP-0356: Privileged Entity

    Events:

    ::

        privileges_advertised  -- Received message/privilege from the server
    """

    name = "xep_0356"
    description = "XEP-0356: Privileged Entity"
    dependencies = {"xep_0297"}
    stanza = stanza

    granted_privileges = defaultdict(Permissions)

    def plugin_init(self):
        if not self.xmpp.is_component:
            log.error("XEP 0356 is only available for components")
            return

        stanza.register()

        self.xmpp.register_handler(
            Callback(
                "Privileges",
                StanzaPath("message/privilege"),
                self._handle_privilege,
            )
        )

    def plugin_end(self):
        self.xmpp.remove_handler("Privileges")

    def _handle_privilege(self, msg: StanzaBase):
        """
        Called when the XMPP server advertise the component's privileges.

        Stores the privileges in this instance's granted_privileges attribute (a dict)
        and raises the privileges_advertised event
        """
        permissions = self.granted_privileges[msg.get_from()]
        for perm in msg["privilege"]["perms"]:
            access = perm["access"]
            if access == "iq":
                for ns in perm["namespaces"]:
                    permissions.iq[ns["ns"]] = ns["type"]
            elif access in _VALID_ACCESSES:
                setattr(permissions, access, perm["type"])
            else:
                log.warning("Received an invalid privileged access: %s", access)
        log.debug(f"Privileges: {self.granted_privileges}")
        self.xmpp.event("privileges_advertised")

    def send_privileged_message(self, msg: Message):
        if (
            self.granted_privileges[msg.get_from().domain].message
            != MessagePermission.OUTGOING
        ):
            raise PermissionError(
                "The server hasn't authorized us to send messages on behalf of other users"
            )
        else:
            self._make_privileged_message(msg).send()

    def _make_privileged_message(self, msg: Message):
        server = msg.get_from().domain
        wrapped = self.xmpp.make_message(mto=server, mfrom=self.xmpp.boundjid.bare)
        wrapped["privilege"]["forwarded"].append(msg)
        return wrapped

    def _make_get_roster(self, jid: typing.Union[JID, str], **iq_kwargs):
        return self.xmpp.make_iq_get(
            queryxmlns="jabber:iq:roster",
            ifrom=self.xmpp.boundjid.bare,
            ito=jid,
            **iq_kwargs,
        )

    def _make_set_roster(
        self,
        jid: typing.Union[JID, str],
        roster_items: dict,
        **iq_kwargs,
    ):
        iq = self.xmpp.make_iq_set(
            ifrom=self.xmpp.boundjid.bare,
            ito=jid,
            **iq_kwargs,
        )
        iq["roster"]["items"] = roster_items
        return iq

    async def get_roster(self, jid: typing.Union[JID, str], **send_kwargs) -> Iq:
        """
        Return the roster of user on the server the component has privileged access to.

        Raises ValueError if the server did not advertise the corresponding privileges

        :param jid: user we want to fetch the roster from
        """
        if isinstance(jid, str):
            jid = JID(jid)
        if self.granted_privileges[jid.domain].roster not in (
            RosterAccess.GET,
            RosterAccess.BOTH,
        ):
            raise PermissionError(
                "The server did not grant us privileges to get rosters"
            )
        else:
            return await self._make_get_roster(jid).send(**send_kwargs)

    async def set_roster(
        self, jid: typing.Union[JID, str], roster_items: dict, **send_kwargs
    ) -> Iq:
        """
        Return the roster of user on the server the component has privileged access to.

        Raises ValueError if the server did not advertise the corresponding privileges

        Here is an example of a roster_items value:

        .. code-block:: json

            {
                "friend1@example.com": {
                    "name": "Friend 1",
                    "subscription": "both",
                    "groups": ["group1", "group2"],
                },
                "friend2@example.com": {
                    "name": "Friend 2",
                    "subscription": "from",
                    "groups": ["group3"],
                },
            }

        :param jid: user we want to add or modify roster items
        :param roster_items: a dict containing the roster items' JIDs as keys and
                             nested dicts containing names, subscriptions and groups.

        """
        if isinstance(jid, str):
            jid = JID(jid)
        if self.granted_privileges[jid.domain].roster not in (
            RosterAccess.GET,
            RosterAccess.BOTH,
        ):
            raise PermissionError(
                "The server did not grant us privileges to set rosters"
            )
        else:
            return await self._make_set_roster(jid, roster_items).send(**send_kwargs)

    async def send_privileged_iq(
        self, encapsulated_iq: Iq, iq_id: typing.Optional[str] = None
    ):
        """
        Send an IQ on behalf of a user

        Caution: the IQ *must* have the jabber:client namespace
        """
        iq_id = iq_id or str(uuid.uuid4())
        encapsulated_iq["id"] = iq_id
        server = encapsulated_iq.get_to().domain
        perms = self.granted_privileges.get(server)
        if not perms:
            raise PermissionError(f"{server} has not granted us any privilege")
        itype = encapsulated_iq["type"]
        for ns in encapsulated_iq.plugins.values():
            type_ = perms.iq[ns.namespace]
            if type_ == IqPermission.NONE:
                raise PermissionError(
                    f"{server} has not granted any IQ privilege for namespace {ns.namespace}"
                )
            elif type_ == IqPermission.BOTH:
                pass
            elif type_ != itype:
                raise PermissionError(
                    f"{server} has not granted IQ {itype} privilege for namespace {ns.namespace}"
                )
        iq = self.xmpp.make_iq(
            itype=itype,
            ifrom=self.xmpp.boundjid.bare,
            ito=encapsulated_iq.get_from(),
            id=iq_id,
        )
        iq["privileged_iq"].append(encapsulated_iq)

        resp = await iq.send()
        return resp["privilege"]["forwarded"]["iq"]


# does not include iq access that is handled differently
_VALID_ACCESSES = {"message", "roster", "presence"}




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Iq
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.google.settings import stanza


class GoogleSettings(BasePlugin):

    """
    Google: Gmail Notifications

    Also see <https://developers.google.com/talk/jep_extensions/usersettings>.
    """

    name = 'google_settings'
    description = 'Google: User Settings'
    dependencies = set()
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, stanza.UserSettings)

        self.xmpp.register_handler(
                Callback('Google Settings',
                    StanzaPath('iq@type=set/google_settings'),
                    self._handle_settings_change))

    def plugin_end(self):
        self.xmpp.remove_handler('Google Settings')

    def get(self, timeout=None, callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq.enable('google_settings')
        return iq.send(timeout=timeout, callback=callback)

    def update(self, settings, timeout=None, callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq.enable('google_settings')

        for setting, value in settings.items():
            iq['google_settings'][setting] = value

        return iq.send(timeout=timeout, callback=callback)

    def _handle_settings_change(self, iq):
        reply = self.xmpp.Iq()
        reply['type'] = 'result'
        reply['id'] = iq['id']
        reply['to'] = iq['from']
        reply.send()
        self.xmpp.event('google_settings_change', iq)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 nicoco
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins import BasePlugin
from . import stanza


class XEP_0492(BasePlugin):
    """
    XEP-0492: Chat notification settings
    """

    name = "xep_0492"
    description = "XEP-0492: Chat notification settings"
    dependencies = {"xep_0402"}
    stanza = stanza

    def plugin_init(self):
        stanza.register_plugin()



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from __future__ import annotations

import logging

from typing import (
    Optional,
    Any,
    Dict,
    Tuple,
    TYPE_CHECKING,
    Union,
    Collection,
)

from slixmpp import BaseXMPP, JID
from slixmpp.stanza import Iq
from slixmpp.types import TypedDict, OptJidStr, OptJid
from slixmpp.exceptions import XMPPError, IqError, IqTimeout
from slixmpp.plugins.xep_0030 import DiscoInfo, DiscoItems


log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from slixmpp.plugins.xep_0030 import XEP_0030


class NodeType(TypedDict):
    info: DiscoInfo
    items: DiscoItems


NodesType = Dict[
    Tuple[str, str, str],
    NodeType
]


class StaticDisco:

    """
    While components will likely require fully dynamic handling
    of service discovery information, most clients and simple bots
    only need to manage a few disco nodes that will remain mostly
    static.

    StaticDisco provides a set of node handlers that will store
    static sets of disco info and items in memory.

    :var nodes: A dictionary mapping (JID, node) tuples to a dict
                containing a disco#info and a disco#items stanza.
    :var xmpp: The main Slixmpp object.
    :var disco: The instance of the XEP-0030 plugin.
    """

    def __init__(self, xmpp: 'BaseXMPP', disco: 'XEP_0030'):
        """
        Create a static disco interface. Sets of disco#info and
        disco#items are maintained for every given JID and node
        combination. These stanzas are used to store disco
        information in memory without any additional processing.

        :param xmpp: The main Slixmpp object.
        :param disco: The XEP-0030 plugin.
        """
        self.nodes: NodesType = {}
        self.xmpp: BaseXMPP = xmpp
        self.disco: 'XEP_0030' = disco

    def add_node(self, jid: OptJidStr = None, node: Optional[str] = None,
                 ifrom: OptJidStr = None) -> NodeType:
        if jid is None:
            node_jid = self.xmpp.boundjid.full
        elif isinstance(jid, JID):
            node_jid = jid.full
        if ifrom is None:
            node_ifrom = ''
        elif isinstance(ifrom, JID):
            node_ifrom = ifrom.full
        else:
            node_ifrom = ifrom
        if node is None:
            node = ''
        if (node_jid, node, node_ifrom) not in self.nodes:
            node_dict: NodeType = {
                'info': DiscoInfo(),
                'items': DiscoItems(),
            }
            node_dict['info']['node'] = node
            node_dict['items']['node'] = node
            self.nodes[(node_jid, node, node_ifrom)] = node_dict
        return self.nodes[(node_jid, node, node_ifrom)]

    def get_node(self, jid: OptJidStr = None, node: Optional[str] = None,
                 ifrom: OptJidStr = None) -> NodeType:
        if jid is None:
            node_jid = self.xmpp.boundjid.full
        elif isinstance(jid, JID):
            node_jid = jid.full
        else:
            node_jid = jid
        if node is None:
            node = ''
        if ifrom is None:
            node_ifrom = ''
        elif isinstance(ifrom, JID):
            node_ifrom = ifrom.full
        else:
            node_ifrom = ifrom
        if (node_jid, node, node_ifrom) not in self.nodes:
            self.add_node(node_jid, node, node_ifrom)
        return self.nodes[(node_jid, node, node_ifrom)]

    def node_exists(self, jid: OptJidStr = None, node: Optional[str] = None,
                    ifrom: OptJidStr = None) -> bool:
        if jid is None:
            node_jid = self.xmpp.boundjid.full
        elif isinstance(jid, JID):
            node_jid = jid.full
        else:
            node_jid = jid
        if node is None:
            node = ''
        if ifrom is None:
            node_ifrom = ''
        elif isinstance(ifrom, JID):
            node_ifrom = ifrom.full
        else:
            node_ifrom = ifrom
        return (node_jid, node, node_ifrom) in self.nodes

    # =================================================================
    # Node Handlers
    #
    # Each handler accepts four arguments: jid, node, ifrom, and data.
    # The jid and node parameters together determine the set of info
    # and items stanzas that will be retrieved or added. Additionally,
    # the ifrom value allows for cached results when results vary based
    # on the requester's JID. The data parameter is a dictionary with
    # additional parameters that will be passed to other calls.
    #
    # This implementation does not allow different responses based on
    # the requester's JID, except for cached results. To do that,
    # register a custom node handler.

    async def supports(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                       data: Any) -> Optional[bool]:
        """
        Check if a JID supports a given feature.

        The data parameter may provide:

        :param feature: The feature to check for support.
        :param local:   If true, then the query is for a JID/node
                        combination handled by this Slixmpp instance and
                        no stanzas need to be sent.
                        Otherwise, a disco stanza must be sent to the
                        remove JID to retrieve the info.
        :param cached:  If true, then look for the disco info data from
                        the local cache system. If no results are found,
                        send the query as usual. The self.use_cache
                        setting must be set to true for this option to
                        be useful. If set to false, then the cache will
                        be skipped, even if a result has already been
                        cached. Defaults to false.
        """
        feature = data.get('feature', None)

        data = {'local': data.get('local', False),
                'cached': data.get('cached', True)}

        if not feature:
            return False

        try:
            info = await self.disco.get_info(jid=jid, node=node,
                                             ifrom=ifrom, **data)
            info = self.disco._wrap(ifrom, jid, info, True)
            features = info['disco_info']['features']
            return feature in features
        except IqError:
            return False
        except IqTimeout:
            return None

    async def has_identity(self, jid: OptJid, node: Optional[str],
                           ifrom: OptJid, data: Dict[str, Any]
                           ) -> Optional[bool]:
        """
        Check if a JID has a given identity.

        The data parameter may provide:

        :param category: The category of the identity to check.
        :param itype:    The type of the identity to check.
        :param lang:     The language of the identity to check.
        :param local:    If true, then the query is for a JID/node
                         combination handled by this Slixmpp instance and
                         no stanzas need to be sent.
                         Otherwise, a disco stanza must be sent to the
                         remove JID to retrieve the info.
        :param cached:   If true, then look for the disco info data from
                         the local cache system. If no results are found,
                         send the query as usual. The self.use_cache
                         setting must be set to true for this option to
                         be useful. If set to false, then the cache will
                         be skipped, even if a result has already been
                         cached. Defaults to false.
        """
        identity = (data.get('category', None),
                    data.get('itype', None),
                    data.get('lang', None))

        data = {'local': data.get('local', False),
                'cached': data.get('cached', True)}

        try:
            info = await self.disco.get_info(jid=jid, node=node,
                                             ifrom=ifrom, **data)
            info = self.disco._wrap(ifrom, jid, info, True)

            def trunc(i):
                return (i[0], i[1], i[2])
            return identity in map(trunc, info['disco_info']['identities'])
        except IqError:
            return False
        except IqTimeout:
            return None

    def get_info(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                 data: Any) -> Optional[DiscoInfo]:
        """
        Return the stored info data for the requested JID/node combination.

        The data parameter is not used.
        """
        if not self.node_exists(jid, node):
            if not node:
                return DiscoInfo()
            else:
                raise XMPPError(condition='item-not-found')
        else:
            return self.get_node(jid, node)['info']

    def set_info(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                 data: DiscoInfo):
        """
        Set the entire info stanza for a JID/node at once.

        The data parameter is a disco#info substanza.
        """
        new_node = self.add_node(jid, node)
        new_node['info'] = data

    def del_info(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                 data: Any):
        """
        Reset the info stanza for a given JID/node combination.

        The data parameter is not used.
        """
        if self.node_exists(jid, node):
            self.get_node(jid, node)['info'] = DiscoInfo()

    def get_items(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                  data: Any) -> Optional[DiscoItems]:
        """
        Return the stored items data for the requested JID/node combination.

        The data parameter is not used.
        """
        if not self.node_exists(jid, node):
            if not node:
                return DiscoItems()
            else:
                raise XMPPError(condition='item-not-found')
        else:
            return self.get_node(jid, node)['items']

    def set_items(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                  data: Dict[str, Collection[Tuple]]):
        """
        Replace the stored items data for a JID/node combination.

        The data parameter may provide:
            items -- A set of items in tuple format.
        """
        items = data.get('items', set())
        new_node = self.add_node(jid, node)
        new_node['items']['items'] = items

    def del_items(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                  data: Any):
        """
        Reset the items stanza for a given JID/node combination.

        The data parameter is not used.
        """
        if self.node_exists(jid, node):
            self.get_node(jid, node)['items'] = DiscoItems()

    def add_identity(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                     data: Dict[str, Optional[str]]):
        """
        Add a new identity to the JID/node combination.

        The data parameter may provide:

        :param category: The general category to which the agent belongs.
        :param itype: A more specific designation with the category.
        :param name: Optional human readable name for this identity.
        :param lang: Optional standard xml:lang value.
        """
        new_node = self.add_node(jid, node)
        new_node['info'].add_identity(
                data.get('category', ''),
                data.get('itype', ''),
                data.get('name', None),
                data.get('lang', None))

    def set_identities(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                       data: Dict[str, Collection[str]]):
        """
        Add or replace all identities for a JID/node combination.

        The data parameter should include:

        :param identities: A list of identities in tuple form:
                           (category, type, name, lang)
        """
        identities = data.get('identities', set())
        new_node = self.add_node(jid, node)
        new_node['info']['identities'] = identities

    def del_identity(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                     data: Dict[str, Optional[str]]):
        """
        Remove an identity from a JID/node combination.

        The data parameter may provide:

        :param category: The general category to which the agent belonged.
        :param itype: A more specific designation with the category.
        :param name: Optional human readable name for this identity.
        :param lang: Optional, standard xml:lang value.
        """
        if self.node_exists(jid, node):
            self.get_node(jid, node)['info'].del_identity(
                    data.get('category', ''),
                    data.get('itype', ''),
                    data.get('name', None),
                    data.get('lang', None))

    def del_identities(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                       data: Any):
        """
        Remove all identities from a JID/node combination.

        The data parameter is not used.
        """
        if self.node_exists(jid, node):
            del self.get_node(jid, node)['info']['identities']

    def add_feature(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                    data: Dict[str, str]):
        """
        Add a feature to a JID/node combination.

        The data parameter should include:

        :param feature: The namespace of the supported feature.
        """
        new_node = self.add_node(jid, node)
        new_node['info'].add_feature(
                data.get('feature', ''))

    def set_features(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                     data: Dict[str, Collection[str]]):
        """
        Add or replace all features for a JID/node combination.

        The data parameter should include:

        :param features: The new set of supported features.
        """
        features = data.get('features', set())
        new_node = self.add_node(jid, node)
        new_node['info']['features'] = features

    def del_feature(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                    data: Dict[str, str]):
        """
        Remove a feature from a JID/node combination.

        The data parameter should include:

        :param feature: The namespace of the removed feature.
        """
        if self.node_exists(jid, node):
            self.get_node(jid, node)['info'].del_feature(
                    data.get('feature', ''))

    def del_features(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                     data: Any):
        """
        Remove all features from a JID/node combination.

        The data parameter is not used.
        """
        if not self.node_exists(jid, node):
            return
        del self.get_node(jid, node)['info']['features']

    def add_item(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                 data: Dict[str, str]):
        """
        Add an item to a JID/node combination.

        The data parameter may include:

        :param ijid: The JID for the item.
        :param inode: Optional additional information to reference
                      non-addressable items.
        :param name: Optional human readable name for the item.
        """
        new_node = self.add_node(jid, node)
        new_node['items'].add_item(
                data.get('ijid', ''),
                node=data.get('inode', ''),
                name=data.get('name', ''))

    def del_item(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                 data: Dict[str, str]):
        """
        Remove an item from a JID/node combination.

        The data parameter may include:

        :param ijid: JID of the item to remove.
        :param inode: Optional extra identifying information.
        """
        if self.node_exists(jid, node):
            self.get_node(jid, node)['items'].del_item(
                    data.get('ijid', ''),
                    node=data.get('inode', None))

    def cache_info(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                   data: Union[Iq, DiscoInfo]):
        """
        Cache disco information for an external JID.

        The data parameter is the Iq result stanza
        containing the disco info to cache, or
        the disco#info substanza itself.
        """
        if isinstance(data, Iq):
            info = data['disco_info']
        else:
            info = data

        new_node = self.add_node(jid, node, ifrom)
        new_node['info'] = info

    def get_cached_info(self, jid: OptJid, node: Optional[str], ifrom: OptJid,
                        data: Any) -> Optional[DiscoInfo]:
        """
        Retrieve cached disco info data.

        The data parameter is not used.
        """
        if not self.node_exists(jid, node, ifrom):
            return None
        return self.get_node(jid, node, ifrom)['info']



from slixmpp.plugins import BasePlugin
from . import stanza


class XEP_0469(BasePlugin):

    """
    XEP-0469: Bookmark Pinning
    """

    name = "xep_0469"
    description = "XEP-0469: Bookmark Pinning"
    dependencies = {"xep_0402"}
    stanza = stanza

    def plugin_init(self):
        stanza.register_plugin()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza.rootstanza import RootStanza
from slixmpp.xmlstream import StanzaBase, ET


ORIGIN_NAME = '{urn:xmpp:sid:0}origin-id'


class Message(RootStanza):

    """
    XMPP's <message> stanzas are a "push" mechanism to send information
    to other XMPP entities without requiring a response.

    Chat clients will typically use <message> stanzas that have a type
    of either "chat" or "groupchat".

    When handling a message event, be sure to check if the message is
    an error response.

    Example <message> stanzas:

    .. code-block:: xml

        <message to="user1@example.com" from="user2@example.com">
          <body>Hi!</body>
        </message>

        <message type="groupchat" to="room@conference.example.com">
          <body>Hi everyone!</body>
        </message>

    Stanza Interface:
        - **body**: The main contents of the message.
        - **subject**: An optional description of the message's contents.
        - **mucroom**: (Read-only) The name of the MUC room that sent the message.
        - **mucnick**: (Read-only) The MUC nickname of message's sender.

    Attributes:
        - **types**: May be one of: normal, chat, headline, groupchat, or error.
    """

    name = 'message'
    namespace = 'jabber:client'
    plugin_attrib = name
    interfaces = {'type', 'to', 'from', 'id', 'body', 'subject', 'thread',
                  'parent_thread', 'mucroom', 'mucnick'}
    sub_interfaces = {'body', 'subject', 'thread'}
    lang_interfaces = sub_interfaces
    types = {'normal', 'chat', 'headline', 'error', 'groupchat'}

    def __init__(self, *args, recv=False, **kwargs):
        """
        Initialize a new <message /> stanza with an optional 'id' value.

        Overrides StanzaBase.__init__.
        """
        StanzaBase.__init__(self, *args, **kwargs)
        if not recv and self['id'] == '':
            if self.stream:
                use_ids = getattr(self.stream, 'use_message_ids', None)
                if use_ids:
                    self.set_id(self.stream.new_id())
            else:
                self.del_origin_id()

    def get_type(self):
        """
        Return the message type.

        Overrides default stanza interface behavior.

        Returns 'normal' if no type attribute is present.

        :rtype: str
        """
        return self._get_attr('type', 'normal')

    def get_id(self):
        return self._get_attr('id') or ''

    def get_origin_id(self):
        sub = self.xml.find(ORIGIN_NAME)
        if sub is not None:
            return sub.attrib.get('id') or ''
        return ''

    def _set_ids(self, value) -> None:
        if value is None or value == '':
            return None

        self.xml.attrib['id'] = value

        if self.stream:
            if not getattr(self.stream, 'use_origin_id', False):
                self.del_origin_id()
                return None

        sub = self.xml.find(ORIGIN_NAME)
        if sub is not None:
            sub.attrib['id'] = value
        else:
            sub = ET.Element(ORIGIN_NAME)
            sub.attrib['id'] = value
            self.xml.append(sub)

    def set_id(self, value):
        return self._set_ids(value)

    def set_origin_id(self, value: str):
        return self._set_ids(value)

    def del_origin_id(self):
        sub = self.xml.find(ORIGIN_NAME)
        if sub is not None:
            self.xml.remove(sub)

    def get_parent_thread(self):
        """Return the message thread's parent thread.

        :rtype: str
        """
        thread = self.xml.find('{%s}thread' % self.namespace)
        if thread is not None:
            return thread.attrib.get('parent', '')
        return ''

    def set_parent_thread(self, value):
        """Add or change the message thread's parent thread.

        :param str value: identifier of the thread"""
        thread = self.xml.find('{%s}thread' % self.namespace)
        if value:
            if thread is None:
                thread = ET.Element('{%s}thread' % self.namespace)
                self.xml.append(thread)
            thread.attrib['parent'] = value
        else:
            if thread is not None and 'parent' in thread.attrib:
                del thread.attrib['parent']

    def del_parent_thread(self):
        """Delete the message thread's parent reference."""
        thread = self.xml.find('{%s}thread' % self.namespace)
        if thread is not None and 'parent' in thread.attrib:
            del thread.attrib['parent']

    def chat(self):
        """Set the message type to 'chat'."""
        self['type'] = 'chat'
        return self

    def normal(self):
        """Set the message type to 'normal'."""
        self['type'] = 'normal'
        return self

    def reply(self, body=None, clear=True):
        """
        Create a message reply.

        Overrides StanzaBase.reply.

        Sets proper 'to' attribute if the message is from a MUC, and
        adds a message body if one is given.

        :param str body:  Optional text content for the message.
        :param bool clear: Indicates if existing content should be removed
                           before replying. Defaults to True.

        :rtype: :class:`~.Message`
        """
        new_message = StanzaBase.reply(self, clear)

        if not getattr(self.stream, "is_component", False) and self['type'] == 'groupchat':
            new_message['to'] = new_message['to'].bare

        new_message['thread'] = self['thread']
        new_message['parent_thread'] = self['parent_thread']

        del new_message['id']
        if self.stream is not None and self.stream.use_message_ids:
            new_message['id'] = self.stream.new_id()

        if body is not None:
            new_message['body'] = body
        return new_message

    def get_mucroom(self):
        """
        Return the name of the MUC room where the message originated.

        Read-only stanza interface.

        :rtype: str
        """
        if self['type'] == 'groupchat':
            return self['from'].bare
        else:
            return ''

    def get_mucnick(self):
        """
        Return the nickname of the MUC user that sent the message.

        Read-only stanza interface.

        :rtype: str
        """
        if self['type'] == 'groupchat':
            return self['from'].resource
        else:
            return ''

    def set_mucroom(self, value):
        """Dummy method to prevent modification."""
        pass

    def del_mucroom(self):
        """Dummy method to prevent deletion."""
        pass

    def set_mucnick(self, value):
        """Dummy method to prevent modification."""
        pass

    def del_mucnick(self):
        """Dummy method to prevent deletion."""
        pass




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from typing import Optional

from slixmpp import JID
from slixmpp.stanza.iq import Iq
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins import xep_0082
from slixmpp.plugins.xep_0202 import stanza


log = logging.getLogger(__name__)


class XEP_0202(BasePlugin):

    """
    XEP-0202: Entity Time
    """

    name = 'xep_0202'
    description = 'XEP-0202: Entity Time'
    dependencies = {'xep_0030', 'xep_0082'}
    stanza = stanza
    default_config = {
        #: As a default, respond to time requests with the
        #: local time returned by XEP-0082. However, a
        #: custom function can be supplied which accepts
        #: the JID of the entity to query for the time.
        'local_time': None,
        'tz_offset': 0
    }

    def plugin_init(self):
        """Start the XEP-0202 plugin."""

        if not self.local_time:
            def default_local_time(jid):
                return xep_0082.datetime(offset=self.tz_offset)

            self.local_time = default_local_time

        self.xmpp.register_handler(
            Callback('Entity Time',
                 StanzaPath('iq@type=get/entity_time'),
                 self._handle_time_request))
        register_stanza_plugin(Iq, stanza.EntityTime)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature='urn:xmpp:time')
        self.xmpp.remove_handler('Entity Time')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('urn:xmpp:time')

    def _handle_time_request(self, iq: Iq):
        """
        Respond to a request for the local time.

        The time is taken from self.local_time(), which may be replaced
        during plugin configuration with a function that maps JIDs to
        times.

        :param iq: The Iq time request stanza.
        """
        iq = iq.reply()
        iq['entity_time']['time'] = self.local_time(iq['to'])
        iq.send()

    def get_entity_time(self, to: JID, ifrom: Optional[JID] = None, **iqargs):
        """
        Request the time from another entity.

        :param to: JID of the entity to query.
        """
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['to'] = to
        iq['from'] = ifrom
        iq.enable('entity_time')
        return iq.send(**iqargs)





# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Message
from slixmpp.util import unicode
from slixmpp.xmlstream import ElementBase, ET, register_stanza_plugin, tostring


XHTML_NS = 'http://www.w3.org/1999/xhtml'


class XHTML_IM(ElementBase):

    namespace = 'http://jabber.org/protocol/xhtml-im'
    name = 'html'
    interfaces = {'body'}
    lang_interfaces = {'body'}
    plugin_attrib = name

    def set_body(self, content, lang=None):
        if lang is None:
            lang = self.get_lang()
        self.del_body(lang)
        if lang == '*':
            for sublang, subcontent in content.items():
                self.set_body(subcontent, sublang)
        else:
            if isinstance(content, type(ET.Element('test'))):
                content = unicode(ET.tostring(content))
            else:
                content = unicode(content)
            header = '<body xmlns="%s"' % XHTML_NS
            if lang:
                header = '%s xml:lang="%s"' % (header, lang)
            content = '%s>%s</body>' % (header, content)
            xhtml = ET.fromstring(content)
            self.xml.append(xhtml)

    def get_body(self, lang=None):
        """Return the contents of the HTML body."""
        if lang is None:
            lang = self.get_lang()

        bodies = self.xml.findall('{%s}body' % XHTML_NS)

        if lang == '*':
            result = {}
            for body in bodies:
                body_lang = body.attrib.get('{%s}lang' % self.xml_ns, '')
                body_result = []
                body_result.append(body.text if body.text else '')
                for child in body:
                    body_result.append(tostring(child, xmlns=XHTML_NS))
                body_result.append(body.tail if body.tail else '')
                result[body_lang] = ''.join(body_result)
            return result
        else:
            for body in bodies:
                if body.attrib.get('{%s}lang' % self.xml_ns, self.get_lang()) == lang:
                    result = []
                    result.append(body.text if body.text else '')
                    for child in body:
                        result.append(tostring(child, xmlns=XHTML_NS))
                    result.append(body.tail if body.tail else '')
                    return ''.join(result)
        return ''

    def del_body(self, lang=None):
        if lang is None:
            lang = self.get_lang()
        bodies = self.xml.findall('{%s}body' % XHTML_NS)
        for body in bodies:
            if body.attrib.get('{%s}lang' % self.xml_ns, self.get_lang()) == lang:
                self.xml.remove(body)
                return




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Dalek
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0249.stanza import Invite
from slixmpp.plugins.xep_0249.invite import XEP_0249


register_plugin(XEP_0249)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase, ET


class Mechanisms(ElementBase):

    """
    """

    name = 'mechanisms'
    namespace = 'urn:ietf:params:xml:ns:xmpp-sasl'
    interfaces = {'mechanisms', 'required'}
    plugin_attrib = name
    is_extension = True

    def get_required(self):
        """
        """
        return True

    def get_mechanisms(self):
        """
        """
        results = []
        mechs = self.xml.findall('{%s}mechanism' % self.namespace)
        if mechs:
            for mech in mechs:
                results.append(mech.text)
        return results

    def set_mechanisms(self, values):
        """
        """
        self.del_mechanisms()
        for val in values:
            mech = ET.Element('{%s}mechanism' % self.namespace)
            mech.text = val
            self.append(mech)

    def del_mechanisms(self):
        """
        """
        mechs = self.xml.findall('{%s}mechanism' % self.namespace)
        if mechs:
            for mech in mechs:
                self.xml.remove(mech)



# slixmpp: The Slick XMPP Library
# Copyright (C) 2018 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.

import logging
import os.path

from aiohttp import ClientSession
from asyncio import Future
from mimetypes import guess_type
from typing import (
    Optional,
    IO,
)

from pathlib import Path

from slixmpp import JID, __version__
from slixmpp.stanza import Iq
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0363 import stanza, Request, Slot, Put, Get, Header

log = logging.getLogger(__name__)

class FileUploadError(Exception):
    pass

class UploadServiceNotFound(FileUploadError):
    """
    Raised if no upload service can be found.
    """

class FileTooBig(FileUploadError):
    """
    Raised if the file size is above advertised server limits.

    args:

    - size of the file
    - max file size allowed
    """
    def __str__(self):
        return 'File size too large: {} (max: {} bytes)' \
            .format(self.args[0], self.args[1])

class HTTPError(FileUploadError):
    """
    Raised when we receive an HTTP error response during upload.

    args:

    - HTTP Error code
    - Content of the HTTP response
    """
    def __str__(self):
        return 'Could not upload file: %d (%s)' % (self.args[0], self.args[1])

class XEP_0363(BasePlugin):
    """
    XEP-0363: HTTP File Upload
    """

    name = 'xep_0363'
    description = 'XEP-0363: HTTP File Upload'
    dependencies = {'xep_0030', 'xep_0128'}
    stanza = stanza
    default_config = {
        'upload_service': None,
        'max_file_size': float('+inf'),
        'default_content_type': 'application/octet-stream',
    }

    def plugin_init(self):
        register_stanza_plugin(Iq, Request)
        register_stanza_plugin(Iq, Slot)
        register_stanza_plugin(Slot, Put)
        register_stanza_plugin(Slot, Get)
        register_stanza_plugin(Put, Header, iterable=True)

        self.xmpp.register_handler(
                Callback('HTTP Upload Request',
                         StanzaPath('iq@type=get/http_upload_request'),
                         self._handle_request))

    def plugin_end(self):
        self.xmpp.remove_handler('HTTP Upload Request')
        self.xmpp['xep_0030'].del_feature(feature=Request.namespace)

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(Request.namespace)

    def _handle_request(self, iq):
        self.xmpp.event('http_upload_request', iq)

    async def find_upload_service(self, domain: Optional[JID] = None, **iqkwargs) -> Optional[Iq]:
        """Find an upload service on a domain (our own by default).

        :param domain: Domain to disco to find a service.
        """
        if domain is None and self.xmpp.is_component:
            domain = self.xmpp.server_host

        results = await self.xmpp['xep_0030'].get_info_from_domain(
            domain=domain, **iqkwargs
        )

        candidates = []
        for info in results:
            if not info['disco_info']:
                continue
            for identity in info['disco_info']['identities']:
                if identity[0] == 'store' and identity[1] == 'file':
                    candidates.append(info)
        for info in candidates:
            for feature in info['disco_info']['features']:
                if feature == Request.namespace:
                    return info

    def request_slot(self, jid: JID, filename: Path, size: int,
                    content_type: Optional[str] = None, *,
                    ifrom: Optional[JID] = None, **iqkwargs) -> Future:
        """Request an HTTP upload slot from a service.

        :param jid: Service to request the slot from.
        :param filename: Name of the file that will be uploaded.
        :param size: size of the file in bytes.
        :param content_type: Type of the file that will be uploaded.
        """
        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        request = iq['http_upload_request']
        request['filename'] = str(filename)
        request['size'] = str(size)
        request['content-type'] = content_type or self.default_content_type
        return iq.send(**iqkwargs)

    async def upload_file(self, filename: Path, size: Optional[int] = None,
                          content_type: Optional[str] = None, *,
                          input_file: Optional[IO[bytes]]=None,
                          domain: Optional[JID] = None,
                          **iqkwargs) -> str:
        '''Helper function which does all of the uploading discovery and
        process.

        :param filename: Path to the file to upload (or only the name if
                         ``input_file`` is provided.
        :param size: size of the file in bytes.
        :param content_type: Type of the file that will be uploaded.
        :param input_file: Binary file stream on the file.
        :param domain: Domain to query to find an HTTP upload service.
        :raises .UploadServiceNotFound: If slixmpp is unable to find an
                                        an available upload service.
        :raises .FileTooBig: If the filesize is above what is accepted by
                             the service.
        :raises .HTTPError: If there is an error in the HTTP operation.
        :returns: The URL of the uploaded file.
        '''
        timeout = iqkwargs.get('timeout', None)
        if self.upload_service is None:
            info_iq = await self.find_upload_service(
                domain=domain, **iqkwargs
            )
            if info_iq is None:
                raise UploadServiceNotFound()
            self.upload_service = info_iq['from']
            for form in info_iq['disco_info'].iterables:
                values = form['values']
                if values['FORM_TYPE'] == ['urn:xmpp:http:upload:0']:
                    try:
                        self.max_file_size = int(values['max-file-size'])
                    except (TypeError, ValueError):
                        log.error('Invalid max size received from HTTP File Upload service')
                        self.max_file_size = float('+inf')
                    break

        if input_file is None:
            input_file = open(filename, 'rb')

        if size is None:
            size = input_file.seek(0, 2)
            input_file.seek(0)

        if size > self.max_file_size:
            raise FileTooBig(size, self.max_file_size)

        if content_type is None:
            content_type = guess_type(filename)[0]
            if content_type is None:
                content_type = self.default_content_type

        basename = os.path.basename(filename)
        slot_iq = await self.request_slot(self.upload_service, basename, size,
                                          content_type, **iqkwargs)
        slot = slot_iq['http_upload_slot']

        headers = {
            'Content-Length': str(size),
            'Content-Type': content_type or self.default_content_type,
            **{header['name']: header['value'] for header in slot['put']['headers']}
        }

        # Do the actual upload here.
        async with ClientSession(headers={'User-Agent': 'slixmpp ' + __version__}) as session:
            response = await session.put(
                    slot['put']['url'],
                    data=input_file,
                    headers=headers,
                    timeout=timeout)
            if response.status >= 400:
                raise HTTPError(response.status, await response.text())
            log.info('Response code: %d (%s)', response.status, await response.text())
            response.close()
            return slot['get']['url']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from base64 import b64encode, b64decode

from slixmpp import JID
from slixmpp.util import bytes
from slixmpp.xmlstream import ElementBase, ET, register_stanza_plugin


class SecurityLabel(ElementBase):
    name = 'securitylabel'
    namespace = 'urn:xmpp:sec-label:0'
    plugin_attrib = 'security_label'

    def add_equivalent(self, label):
        equiv = EquivalentLabel(parent=self)
        equiv.append(label)
        return equiv


class Label(ElementBase):
    name = 'label'
    namespace = 'urn:xmpp:sec-label:0'
    plugin_attrib = 'label'


class DisplayMarking(ElementBase):
    name = 'displaymarking'
    namespace = 'urn:xmpp:sec-label:0'
    plugin_attrib = 'display_marking'
    interfaces = {'fgcolor', 'bgcolor', 'value'}

    def get_fgcolor(self):
        return self._get_attr('fgcolor', 'black')

    def get_bgcolor(self):
        return self._get_attr('fgcolor', 'white')

    def get_value(self):
        return self.xml.text

    def set_value(self, value):
        self.xml.text = value

    def del_value(self):
        self.xml.text = ''


class EquivalentLabel(ElementBase):
    name = 'equivalentlabel'
    namespace = 'urn:xmpp:sec-label:0'
    plugin_attrib = 'equivalent_label'
    plugin_multi_attrib = 'equivalent_labels'


class Catalog(ElementBase):
    name = 'catalog'
    namespace = 'urn:xmpp:sec-label:catalog:2'
    plugin_attrib = 'security_label_catalog'
    interfaces = {'to', 'from', 'name', 'desc', 'id', 'size', 'restrict'}

    def get_to(self):
        return JID(self._get_attr('to'))
        pass

    def set_to(self, value):
        return self._set_attr('to', str(value))

    def get_from(self):
        return JID(self._get_attr('from'))

    def set_from(self, value):
        return self._set_attr('from', str(value))

    def get_restrict(self):
        value = self._get_attr('restrict', '')
        if value and value.lower() in ('true', '1'):
            return True
        return False

    def set_restrict(self, value):
        self._del_attr('restrict')
        if value:
            self._set_attr('restrict', 'true')
        elif value is False:
            self._set_attr('restrict', 'false')


class CatalogItem(ElementBase):
    name = 'catalog'
    namespace = 'urn:xmpp:sec-label:catalog:2'
    plugin_attrib = 'item'
    plugin_multi_attrib = 'items'
    interfaces = {'selector', 'default'}

    def get_default(self):
        value = self._get_attr('default', '')
        if value.lower() in ('true', '1'):
            return True
        return False

    def set_default(self, value):
        self._del_attr('default')
        if value:
            self._set_attr('default', 'true')
        elif value is False:
            self._set_attr('default', 'false')


class ESSLabel(ElementBase):
    name = 'esssecuritylabel'
    namespace = 'urn:xmpp:sec-label:ess:0'
    plugin_attrib = 'ess'
    interfaces = {'value'}

    def get_value(self):
        if self.xml.text:
            return b64decode(bytes(self.xml.text))
        return ''

    def set_value(self, value):
        self.xml.text = ''
        if value:
            self.xml.text = b64encode(bytes(value))

    def del_value(self):
        self.xml.text = ''


register_stanza_plugin(Catalog, CatalogItem, iterable=True)
register_stanza_plugin(CatalogItem, SecurityLabel)
register_stanza_plugin(EquivalentLabel, ESSLabel)
register_stanza_plugin(Label, ESSLabel)
register_stanza_plugin(SecurityLabel, DisplayMarking)
register_stanza_plugin(SecurityLabel, EquivalentLabel, iterable=True)
register_stanza_plugin(SecurityLabel, Label)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0016 import stanza
from slixmpp.plugins.xep_0016.stanza import Privacy
from slixmpp.plugins.xep_0016.privacy import XEP_0016


register_plugin(XEP_0016)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8 et ts=4 sts=4 sw=4
#
# Copyright © 2022 Maxime “pep” Buquet <pep@bouah.net>
#
# See the LICENSE file for copying permissions.

"""
    XEP-0454: OMEMO Media Sharing
"""

from typing import IO, Optional, Tuple

from os import urandom
from pathlib import Path
from io import BytesIO, SEEK_END

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from slixmpp.plugins import BasePlugin
from slixmpp.plugins.base import register_plugin


class InvalidURL(Exception):
    """Raised for URLs that either aren't HTTPS or already contain a fragment."""


EXTENSIONS_MAP = {
    'jpeg': 'jpg',
    'text': 'txt',
}

class XEP_0454(BasePlugin):
    """
        XEP-0454: OMEMO Media Sharing
    """

    name = 'xep_0454'
    description = 'XEP-0454: OMEMO Media Sharing'
    dependencies = {'xep_0363'}

    @staticmethod
    def encrypt(input_file: Optional[IO[bytes]] = None, filename: Optional[Path] = None) -> Tuple[bytes, str]:
        """
            Encrypts file as specified in XEP-0454 for use in file sharing

            :param input_file: Binary file stream on the file.
            :param filename: Path to the file to upload.

            One of input_file or filename must be specified. If both are
            passed, input_file will be used and filename ignored.
        """
        if input_file is None and filename is None:
            raise ValueError('Specify either filename or input_file parameter')

        aes_gcm_iv = urandom(12)
        aes_gcm_key = urandom(32)

        aes_gcm = Cipher(
            algorithms.AES(aes_gcm_key),
            modes.GCM(aes_gcm_iv),
        ).encryptor()

        if input_file is None:
            input_file = open(filename, 'rb')

        payload = b''
        while True:
            buf = input_file.read(4096)
            if not buf:
                break
            payload += aes_gcm.update(buf)

        payload += aes_gcm.finalize() + aes_gcm.tag
        fragment = aes_gcm_iv.hex() + aes_gcm_key.hex()
        return (payload, fragment)

    @staticmethod
    def decrypt(input_file: IO[bytes], fragment: str) -> bytes:
        """
            Decrypts file-like.

            :param input_file: Binary file stream on the file, containing the
                               tag (16 bytes) at the end.
            :param fragment: 88 hex chars string composed of iv (24 chars)
                             + key (64 chars).
        """

        assert len(fragment) == 88
        aes_gcm_iv = bytes.fromhex(fragment[:24])
        aes_gcm_key = bytes.fromhex(fragment[24:])

        # Find 16 bytes tag
        input_file.seek(-16, SEEK_END)
        tag = input_file.read()

        aes_gcm = Cipher(
            algorithms.AES(aes_gcm_key),
            modes.GCM(aes_gcm_iv, tag),
        ).decryptor()

        size = input_file.seek(0, SEEK_END)
        input_file.seek(0)

        count = size - 16
        plain = b''
        while count >= 0:
            buf = input_file.read(4096)
            count -= len(buf)
            if count <= 0:
                buf += input_file.read()
                buf = buf[:-16]
            plain += aes_gcm.update(buf)
        plain += aes_gcm.finalize()

        return plain

    @staticmethod
    def format_url(url: str, fragment: str) -> str:
        """Helper to format a HTTPS URL to an AESGCM URI"""
        if not url.startswith('https://') or url.find('#') != -1:
            raise InvalidURL
        return 'aesgcm://' + url[len('https://'):] + '#' + fragment

    @staticmethod
    def map_extensions(ext: str) -> str:
        """
            Apply conversions to extensions to reduce the number of
            variations, (e.g., JPEG -> jpg).
        """
        return EXTENSIONS_MAP.get(ext, ext).lower()

    async def upload_file(
        self,
        filename: Path,
        _size: Optional[int] = None,
        content_type: Optional[str] = None,
        **kwargs,
    ) -> str:
        """
            Wrapper to xep_0363 (HTTP Upload)'s upload_file method.

            :param input_file: Binary file stream on the file.
            :param filename: Path to the file to upload.

            Same as `XEP_0454.encrypt`, one of input_file or filename must be
            specified. If both are passed, input_file will be used and
            filename ignored.

            Other arguments passed in are passed to the actual
            `XEP_0363.upload_file` call.
        """
        input_file = kwargs.get('input_file')
        payload, fragment = self.encrypt(input_file, filename)

        # Prepare kwargs for upload_file call
        new_filename = urandom(12).hex()  # Random filename to hide user-provided path
        if filename.suffix:
            new_filename += self.map_extensions(filename.suffix)
        kwargs['filename'] = new_filename

        input_enc = BytesIO(payload)
        kwargs['input_file'] = input_enc

        # Size must also be overriden if provided
        size = input_enc.seek(0, SEEK_END)
        input_enc.seek(0)
        kwargs['size'] = size

        kwargs['content_type'] = content_type

        url = await self.xmpp['xep_0363'].upload_file(**kwargs)
        return self.format_url(url, fragment)

register_plugin(XEP_0454)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.jid import JID
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class NoSave(ElementBase):
    name = 'x'
    namespace = 'google:nosave'
    plugin_attrib = 'google_nosave'
    interfaces = {'value'}

    def get_value(self):
        return self._get_attr('value', '') == 'enabled'

    def set_value(self, value):
        self._set_attr('value', 'enabled' if value else 'disabled')


class NoSaveQuery(ElementBase):
    name = 'query'
    namespace = 'google:nosave'
    plugin_attrib = 'google_nosave'
    interfaces = set()


class Item(ElementBase):
    name = 'item'
    namespace = 'google:nosave'
    plugin_attrib = 'item'
    plugin_multi_attrib = 'items'
    interfaces = {'jid', 'source', 'value'}

    def get_value(self):
        return self._get_attr('value', '') == 'enabled'

    def set_value(self, value):
        self._set_attr('value', 'enabled' if value else 'disabled')

    def get_jid(self):
        return JID(self._get_attr('jid', ''))

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_source(self):
        return JID(self._get_attr('source', ''))

    def set_source(self, value):
        self._set_attr('source', str(value))


register_stanza_plugin(NoSaveQuery, Item)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import unittest
from slixmpp import Message
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0444 import XEP_0444
import slixmpp.plugins.xep_0444.stanza as stanza
from slixmpp.xmlstream import register_stanza_plugin

try:
    import emoji
except ImportError:
    emoji = None


class TestReactions(SlixTest):

    def setUp(self):
        register_stanza_plugin(Message, stanza.Reactions)
        register_stanza_plugin(stanza.Reactions, stanza.Reaction, iterable=True)

    def testCreateReactions(self):
        """Testing creating Reactions."""

        xmlstring = """
          <message>
              <reactions xmlns="urn:xmpp:reactions:0" id="abcd">
                  <reaction>😃</reaction>
                  <reaction>🤗</reaction>
              </reactions>
          </message>
        """

        msg = self.Message()
        msg['reactions']['id'] = 'abcd'
        msg['reactions']['values'] = ['😃', '🤗']

        self.check(msg, xmlstring, use_values=False)

        self.assertEqual({'😃', '🤗'}, msg['reactions']['values'])


    @unittest.skipIf(emoji is None, 'Emoji package not installed')
    def testCreateReactionsUnrestricted(self):
        """Testing creating Reactions with the extra all_chars arg."""
        xmlstring = """
          <message>
              <reactions xmlns="urn:xmpp:reactions:0" id="abcd">
                  <reaction>😃</reaction>
                  <reaction>🤗</reaction>
                  <reaction>toto</reaction>
              </reactions>
          </message>
        """

        msg = self.Message()
        msg['reactions']['id'] = 'abcd'
        msg['reactions'].set_values(['😃', '🤗', 'toto'], all_chars=True)

        self.check(msg, xmlstring, use_values=False)

        self.assertEqual({'😃', '🤗'}, msg['reactions']['values'])
        self.assertEqual({'😃', '🤗', 'toto'}, msg['reactions'].get_values(all_chars=True))
        with self.assertRaises(ValueError):
            msg['reactions'].set_values(['😃', '🤗', 'toto'], all_chars=False)


suite = unittest.TestLoader().loadTestsFromTestCase(TestReactions)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import datetime as dt

from slixmpp import Message
from slixmpp.xmlstream import register_stanza_plugin, ElementBase, ET, JID
from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins import xep_0082


class Event(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'event'
    plugin_attrib = 'pubsub_event'
    interfaces = set()


class EventItem(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'item'
    plugin_attrib = name
    interfaces = {'id', 'payload', 'node', 'publisher'}

    def set_payload(self, value):
        self.xml.append(value)

    def get_payload(self):
        children = list(self.xml)
        if len(children) > 0:
            return children[0]

    def del_payload(self):
        for child in self.xml:
            self.xml.remove(child)


class EventRetract(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'retract'
    plugin_attrib = name
    interfaces = {'id'}


class EventItems(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'items'
    plugin_attrib = name
    interfaces = {'node'}


class EventCollection(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'collection'
    plugin_attrib = name
    interfaces = {'node'}


class EventAssociate(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'associate'
    plugin_attrib = name
    interfaces = {'node'}


class EventDisassociate(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'disassociate'
    plugin_attrib = name
    interfaces = {'node'}


class EventConfiguration(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'configuration'
    plugin_attrib = name
    interfaces = {'node'}


class EventPurge(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'purge'
    plugin_attrib = name
    interfaces = {'node'}


class EventDelete(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'delete'
    plugin_attrib = name
    interfaces = {'node', 'redirect'}

    def set_redirect(self, uri):
        del self['redirect']
        redirect = ET.Element('{%s}redirect' % self.namespace)
        redirect.attrib['uri'] = uri
        self.xml.append(redirect)

    def get_redirect(self):
        redirect = self.xml.find('{%s}redirect' % self.namespace)
        if redirect is not None:
            return redirect.attrib.get('uri', '')
        return ''

    def del_redirect(self):
        redirect = self.xml.find('{%s}redirect' % self.namespace)
        if redirect is not None:
            self.xml.remove(redirect)


class EventSubscription(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#event'
    name = 'subscription'
    plugin_attrib = name
    interfaces = {'node', 'expiry', 'jid', 'subid', 'subscription'}

    def get_expiry(self):
        expiry = self._get_attr('expiry')
        if expiry.lower() == 'presence':
            return expiry
        return xep_0082.parse(expiry)

    def set_expiry(self, value):
        if isinstance(value, dt.datetime):
            value = xep_0082.format_datetime(value)
        self._set_attr('expiry', value)

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


register_stanza_plugin(Message, Event)
register_stanza_plugin(Event, EventCollection)
register_stanza_plugin(Event, EventConfiguration)
register_stanza_plugin(Event, EventPurge)
register_stanza_plugin(Event, EventDelete)
register_stanza_plugin(Event, EventItems)
register_stanza_plugin(Event, EventSubscription)
register_stanza_plugin(EventCollection, EventAssociate)
register_stanza_plugin(EventCollection, EventDisassociate)
register_stanza_plugin(EventConfiguration, Form)
register_stanza_plugin(EventItems, EventItem, iterable=True)
register_stanza_plugin(EventItems, EventRetract, iterable=True)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission
import logging
from asyncio import Event, Task, FIRST_COMPLETED, wait, CancelledError
from enum import Enum
from typing import Optional, Tuple, Union

from slixmpp.stanza import Message, Presence
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.jid import JID
from slixmpp.plugins import BasePlugin
from slixmpp.types import ErrorConditions


log = logging.getLogger(__name__)


class PingStatus(Enum):
    """
    Enum representing the status of a ping.

    - `UNTRIED`: the ping was not sent yet
    - `JOINED`: we were considered still in the MUC at the last ping
    - `JOINED_NICKCHANGED`: we were joined but another client just changed
      the nick.
    - `DISCONNECTED`: we got disconnected
    - `TIMEOUT`: slixmpp was unable to reach the server
    """
    UNTRIED = 0
    JOINED = 1
    JOINED_NICKCHANGED = 1
    DISCONNECTED = 2
    TIMEOUT = 3


class PingTask:
    """
    Class representing a current muc self-ping task.
    """
    _event: Event
    _current_task: Task
    _timeout: float
    _interval: float
    _plugin: 'XEP_0410'

    def __init__(self, muc_resource: JID, orig_jid: JID,
                 plugin: 'XEP_0410', interval: float, timeout: float) -> None:
        self._event = Event()
        self._plugin = plugin
        self._interval = interval
        self._timeout = timeout
        self._current_task = plugin.xmpp.loop.create_task(self.run(
            muc_resource,
            orig_jid,
        ))

    def reset_timer(self) -> None:
        """
        Triggers the internal event to reset the internal timer.
        """
        self._event.set()

    def cancel(self) -> None:
        """Cancel the task before dying"""
        self._current_task.cancel()

    async def run(self, muc_resource: JID, orig_jid: JID) -> None:
        """
        Loop that goes on forever and sends pings at the desired interval.
        """
        while True:
            # Wait on the event or timeout, whichever comes first
            try:
                done, pending = await wait(
                    [self._plugin.xmpp.loop.create_task(self._event.wait())],
                    return_when=FIRST_COMPLETED,
                    timeout=self._interval,
                )
                # If the event is set, then the timer was reset and we clear it
                # before going back to waiting
                if self._event.is_set():
                    self._event.clear()
                    continue
                result = await self._plugin.send_self_ping(
                    muc_resource,
                    orig_jid,
                    timeout=self._timeout,
                )
                key = (muc_resource, orig_jid)
                self._plugin._update_ping_results(key, result)
            except CancelledError:
                return
            except:
                log.exception("Error while trying to ping a muc resource")
                return


class XEP_0410(BasePlugin):

    """
    XEP-0410: MUC Self-Ping (Schrödinger's Chat)

    This plugin provides features for clients (or non-MUC components) to
    enable or disable self-ping.

    Configuration options:

    - `ping_interval`: the number of seconds between the last activity and a ping.

    Ping status is represented using the :class:`~slixmpp.plugins.XEP_0410.PingStatus` class.

    Users of this plugin can either schedule the pings manually and send them
    using `send_self_ping`, or they can use the provided scheduler with
    `enable_self_ping` and `disable_self_ping`.

    Doing so requires calling `update_nick` if your nickname in a MUC changes,
    because the plugin is not aware of this. The scheduled ping will trigger
    a `muc_ping_changed` event in case the ping status changes, which should
    be handled properly.
    """

    name = 'xep_0410'
    description = 'XEP-0410: MUC Self-Ping (Schrödinger\'s Chat)'
    dependencies = {'xep_0045', 'xep_0199'}
    default_config = {
        "ping_interval": 900,
        "ping_timeout": 30,
    }
    ping_interval: Union[int, float]
    # Cache of the last bound JID, to be able to recover if we bind to
    # another resource while running
    boundjid: Optional[JID] = None
    # Dictionary mapping a (muc resource, from jid) to an asyncio task in
    # the process of being executed
    ping_timers: dict[Tuple[JID, JID], PingTask]
    # Cache of the latest ping results
    last_ping_results: dict[Tuple[JID, JID], PingStatus]


    def plugin_init(self):
        self.ping_timers = dict()
        self.boundjid = None
        self.last_ping_results = dict()
        self.xmpp.add_event_handler(
            'groupchat_message',
            self._on_muc_activity
        )
        self.xmpp.add_event_handler(
            'groupchat_presence',
            self._on_muc_activity,
        )

    def plugin_end(self):
        self.ping_timers = dict()
        self.last_ping_results = dict()

    def session_bind(self, jid: JID):
        self.__update_boundjid(jid)

    def __update_boundjid(self, jid: JID):
        """
        If the bound JID has changed, we need to update the timers so that
        the ping stanzas have the correct from value.

        This is a bit painful.
        """
        new_boundjid = jid
        if self.boundjid and self.boundjid != new_boundjid:
            self.ping_results = dict()
            for key, timer in list(self.ping_timers.items()):
                if key[1] == self.boundjid:
                    del self.ping_timers[key]
                    timer.cancel()
                    key = (key[0], new_boundjid)
                    self.ping_timers[key] = PingTask(
                        key[0],
                        key[1],
                        self,
                        timeout=timer.timeout,
                        interval=timer.interval,
                    )
        self.boundjid = jid

    def update_nick(self, previous_jid: JID, new_jid: JID) -> None:
        """
        Update the self-ping targets when a nickname changes.

        :param previous_jid: full JID of the previous MUC resource.
        :param new_jid: full JID of the new MUC resource.
        """
        for key, timer in list(self.ping_timers.items()):
            if key[0] == previous_jid:
                del self.ping_timers[key]
                timer.cancel()
                key = (new_jid, key[1])
                self.ping_timers[key] = PingTask(
                    key[0],
                    key[1],
                    self,
                    timeout=timer.timeout,
                    interval=timer.interval,
                )

    def enable_self_ping(self, muc_resource: JID,
                         orig_jid: Optional[JID] = None,
                         interval: Optional[float] = None,
                         timeout: Optional[float] = None) -> None:
        """
        Enable client self-ping.
        The given MUC resource will be pinged periodically if the MUC is inactive,
        and an event will be generated in case of changes.

        :param muc_resource: The MUC resource to send the ping to.
        :param orig_jid: The "from" resource to send the ping from (for components).
        """
        if orig_jid is None:
            orig_jid = self.xmpp.boundjid
        if timeout is None:
            timeout = self.ping_timeout
        if interval is None:
            interval = self.ping_interval
        key = (muc_resource, orig_jid)
        if key not in self.ping_timers:
            self.ping_timers[key] = PingTask(
                muc_resource=muc_resource,
                orig_jid=orig_jid,
                plugin=self,
                timeout=timeout,
                interval=interval,
            )

    def disable_self_ping(self, muc_resource: JID,
                                 orig_jid: Optional[JID] = None) -> None:
        """
        Disable client self-ping. Cancels the scheduled pings for the given
        MUC resource.

        :param muc_resource: The MUC resource to send the ping to.
        :param orig_jid: The "from" resource the ping was sent from (for components).
        """
        if orig_jid is None:
            orig_jid = self.xmpp.boundjid
        key = (muc_resource, orig_jid)
        if key in self.ping_timers:
            self.ping_timers[key].cancel()
            del self.ping_timers[key]
        if key in self.last_ping_results:
            del self.last_ping_results[key]

    def get_ping_status(self, muc_resource: JID,
                        orig_jid: Optional[JID] = None) -> PingStatus:
        """
        Return the last pinged status for a specific muc resource.

        :param muc_resource: The MUC resource to send the ping to.
        :param orig_jid: The "from" resource the ping was sent from (for components).
        :return: The status, or None if no ping has been sent yet.
        """
        if orig_jid is None:
            orig_jid = self.xmpp.boundjid
        key = (muc_resource, orig_jid)
        return self.last_ping_results.get(key, PingStatus.UNTRIED)

    def _on_muc_activity(self, event: Union[Presence, Message]):
        """Handle both messages and presences from mucs to see if we need to
        reset the timer"""
        if event['type'] == 'error':
            return
        key = (event['from'], event['to'])
        if key in self.ping_timers:
            self.ping_timers[key].reset_timer()

    def _handle_condition(self, condition: ErrorConditions) -> PingStatus:
        """
        Interpret the error conditions as defined in XEP-0410.
        """
        if condition in ('service-unavailable', 'feature-not-implemented'):
            return PingStatus.JOINED
        elif condition == 'item-not-found':
            return PingStatus.JOINED_NICKCHANGED
        elif condition in ('remote-server-not-found',
                           'remote-server-timeout'):
            return PingStatus.TIMEOUT
        else:
            return PingStatus.DISCONNECTED

    async def send_self_ping(self, muc_resource: JID,
                             orig_jid: Optional[JID] = None,
                             timeout: Optional[float] = None) -> PingStatus:
        """
        Send a single self-ping to a MUC, and return the result.

        :param muc_resource: The MUC resource to send the ping to.
        :param orig_jid: The "from" resource the ping was sent from (for components).
        :return: The ping status
        """
        if orig_jid is None:
            orig_jid = self.xmpp.boundjid
        result = PingStatus.UNTRIED
        try:
            await self.xmpp.plugin['xep_0199'].send_ping(
                muc_resource,
                ifrom=orig_jid,
                timeout=timeout,
            )
            result = PingStatus.JOINED
        except IqTimeout:
            result = PingStatus.TIMEOUT
        except IqError as exc:
            result = self._handle_condition(exc.condition)
        return result

    def _update_ping_results(self, key: Tuple[JID, JID], result: PingStatus) -> None:
        """
        Internal use only: used to update the ping results dict from the timer.
        """
        previous = self.last_ping_results.get(key, PingStatus.UNTRIED)
        self.last_ping_results[key] = result
        if result != previous:
            self.xmpp.event('muc_ping_changed', {
                "key": key,
                "previous": previous,
                "result": result
            })



import unittest
import slixmpp
from slixmpp.test import SlixTest

from slixmpp.plugins.xep_0172 import UserNick
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.stanza import Presence

register_stanza_plugin(Presence, UserNick)

class TestPresenceStanzas(SlixTest):

    def testPresenceShowRegression(self):
        """Regression check presence['type'] = 'dnd' show value working"""
        p = self.Presence()
        p['type'] = 'dnd'
        self.check(p, "<presence><show>dnd</show></presence>")

    def testPresenceType(self):
        """Test manipulating presence['type']"""
        p = self.Presence()
        p['type'] = 'available'
        self.check(p, "<presence />")
        self.assertTrue(p['type'] == 'available',
            "Incorrect presence['type'] for type 'available': %s" % p['type'])

        for showtype in ['away', 'chat', 'dnd', 'xa']:
            p['type'] = showtype
            self.check(p, """
              <presence><show>%s</show></presence>
            """ % showtype)
            self.assertTrue(p['type'] == showtype,
                "Incorrect presence['type'] for type '%s'" % showtype)

        p['type'] = None
        self.check(p, "<presence />")

    def testPresenceUnsolicitedOffline(self):
        """
        Unsolicted offline presence does not spawn changed_status
        or update the roster.
        """
        p = self.Presence()
        p['type'] = 'unavailable'
        p['from'] = 'bill@chadmore.com/gmail15af'

        c = slixmpp.ClientXMPP('crap@wherever', 'password')
        happened = []

        def handlechangedpresence(event):
            happened.append(True)

        c.add_event_handler("changed_status", handlechangedpresence)
        c._handle_presence(p)

        self.assertTrue(happened == [],
            "changed_status event triggered for extra unavailable presence")
        roster = c.roster['crap@wherever']
        self.assertTrue(roster['bill@chadmore.com'].resources == {},
            "Roster updated for superfulous unavailable presence")

    def testNickPlugin(self):
        """Test presence/nick/nick stanza."""
        p = self.Presence()
        p['nick']['nick'] = 'A nickname!'
        self.check(p, """
          <presence>
            <nick xmlns="http://jabber.org/protocol/nick">A nickname!</nick>
          </presence>
        """)


suite = unittest.TestLoader().loadTestsFromTestCase(TestPresenceStanzas)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0334 import stanza, Store, NoStore, NoPermanentStore, NoCopy

log = logging.getLogger(__name__)

class XEP_0334(BasePlugin):

    name = 'xep_0334'
    description = 'XEP-0334: Message Processing Hints'
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, Store)
        register_stanza_plugin(Message, NoStore)
        register_stanza_plugin(Message, NoPermanentStore)
        register_stanza_plugin(Message, NoCopy)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.features.feature_mechanisms.mechanisms import FeatureMechanisms
from slixmpp.features.feature_mechanisms.stanza import Mechanisms
from slixmpp.features.feature_mechanisms.stanza import Auth
from slixmpp.features.feature_mechanisms.stanza import Success
from slixmpp.features.feature_mechanisms.stanza import Failure


register_plugin(FeatureMechanisms)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2021 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0382.stanza import *
from slixmpp.plugins.xep_0382.spoiler import XEP_0382

register_plugin(XEP_0382)




# slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from typing import Iterable, List, Tuple

from slixmpp.xmlstream import ElementBase, ET

class JingleMessage(ElementBase):
    namespace = 'urn:xmpp:jingle-message:0'
    interfaces = {'id'}

class Propose(JingleMessage):
    name = 'propose'
    plugin_attrib = 'jingle_propose'
    interfaces = {'id', 'descriptions'}

    def get_descriptions(self) -> List[Tuple[str, str]]:
        result = []
        for desc in self.xml:
            namespace = desc.tag.split('}')[0][1:]
            media = desc.attrib['media']
            result.append((namespace, media))
        return result

    def set_descriptions(self, descriptions: Iterable[Tuple[str, str]]):
        self.del_descriptions()
        for namespace, media in descriptions:
            desc = ET.Element('{%s}description' % namespace)
            desc.attrib['media'] = media
            self.xml.append(desc)

    def del_descriptions(self):
        for desc in self.xml.findall('{*}description'):
            self.xml.remove(desc)

class Retract(JingleMessage):
    name = 'retract'
    plugin_attrib = 'jingle_retract'

class Accept(JingleMessage):
    name = 'accept'
    plugin_attrib = 'jingle_accept'

class Proceed(JingleMessage):
    name = 'proceed'
    plugin_attrib = 'jingle_proceed'

class Reject(JingleMessage):
    name = 'reject'
    plugin_attrib = 'jingle_reject'



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp
from slixmpp.exceptions import IqError, IqTimeout


class Disco(slixmpp.ClientXMPP):

    """
    A demonstration for using basic service discovery.

    Send a disco#info and disco#items request to a JID/node combination,
    and print out the results.

    May also request only particular info categories such as just features,
    or just items.
    """

    def __init__(self, jid, password, target_jid, target_node='', get=''):
        slixmpp.ClientXMPP.__init__(self, jid, password)

        # Using service discovery requires the XEP-0030 plugin.
        self.register_plugin('xep_0030')

        self.get = get
        self.target_jid = target_jid
        self.target_node = target_node

        # Values to control which disco entities are reported
        self.info_types = ['', 'all', 'info', 'identities', 'features']
        self.identity_types = ['', 'all', 'info', 'identities']
        self.feature_types = ['', 'all', 'info', 'features']
        self.items_types = ['', 'all', 'items']


        # The session_start event will be triggered when
        # the bot establishes its connection with the server
        # and the XML streams are ready for use. We want to
        # listen for this event so that we we can initialize
        # our roster.
        self.add_event_handler("session_start", self.start)

    async def start(self, event):
        """
        Process the session_start event.

        Typical actions for the session_start event are
        requesting the roster and broadcasting an initial
        presence stanza.

        In this case, we send disco#info and disco#items
        stanzas to the requested JID and print the results.

        Arguments:
            event -- An empty dictionary. The session_start
                     event does not provide any additional
                     data.
        """
        await self.get_roster()
        self.send_presence()

        try:
            if self.get in self.info_types:
                # function using the callback parameter.
                info = await self['xep_0030'].get_info(jid=self.target_jid,
                                                            node=self.target_node)
            if self.get in self.items_types:
                # The same applies from above. Listen for the
                # disco_items event or pass a callback function
                # if you need to process a non-blocking request.
                items = await self['xep_0030'].get_items(jid=self.target_jid,
                                                              node=self.target_node)
            if self.get not in self.info_types and self.get not in self.items_types:
                logging.error("Invalid disco request type.")
                return
        except IqError as e:
            logging.error("Entity returned an error: %s" % e.iq['error']['condition'])
        except IqTimeout:
            logging.error("No response received.")
        else:
            header = 'XMPP Service Discovery: %s' % self.target_jid
            print(header)
            print('-' * len(header))
            if self.target_node != '':
                print('Node: %s' % self.target_node)
                print('-' * len(header))

            if self.get in self.identity_types:
                print('Identities:')
                for identity in info['disco_info']['identities']:
                    print('  - %s' % str(identity))

            if self.get in self.feature_types:
                print('Features:')
                for feature in info['disco_info']['features']:
                    print('  - %s' % feature)

            if self.get in self.items_types:
                print('Items:')
                for item in items['disco_items']['items']:
                    print('  - %s' % str(item))
        finally:
            self.disconnect()


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser(description=Disco.__doc__)

    parser.add_argument("-q","--quiet", help="set logging to ERROR",
                        action="store_const",
                        dest="loglevel",
                        const=logging.ERROR,
                        default=logging.ERROR)
    parser.add_argument("-d","--debug", help="set logging to DEBUG",
                        action="store_const",
                        dest="loglevel",
                        const=logging.DEBUG,
                        default=logging.ERROR)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")
    parser.add_argument("query", choices=["all", "info", "items", "identities", "features"])
    parser.add_argument("target_jid")
    parser.add_argument("node", nargs='?')

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")

    # Setup the Disco browser.
    xmpp = Disco(args.jid, args.password, args.target_jid, args.node, args.query)

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_until_complete(xmpp.disconnected)



import logging
import unittest
from slixmpp.test import SlixTest


class TestReply(SlixTest):
    def setUp(self):
        self.stream_start(plugins=["xep_0461"])

    def testFallBackBody(self):
        def on_reply(msg):
            start = msg["fallback"]["body"]["start"]
            end = msg["fallback"]["body"]["end"]
            self.xmpp["xep_0461"].send_reply(
                reply_to=msg.get_from(),
                reply_id=msg.get_id(),
                fallback=msg["reply"].strip_fallback_content(),
                quoted_nick="res",
                mto="test@test.com",
                mbody=f"{start} to {end}",
            )

        self.xmpp.add_event_handler("message_reply", on_reply)

        self.recv(
            """
            <message id="other-id" from="from@from.com/res">
              <reply xmlns="urn:xmpp:reply:0" id="some-id" />
              <body>&gt; quoted\nsome-body</body>
                <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:reply:0'>
                   <body start="0" end="9" />
                </fallback>
            </message>
            """
        )
        self.send(
            """
            <message xmlns="jabber:client" to="test@test.com" type="normal">
              <body>&gt; res:\n&gt; some-body\n0 to 9</body>
              <reply xmlns="urn:xmpp:reply:0" id="other-id" to="from@from.com/res" />
              <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:reply:0'>
                <body start="0" end="19" />
              </fallback>
            </message>
            """
        )


logging.basicConfig(level=logging.DEBUG)
suite = unittest.TestLoader().loadTestsFromTestCase(TestReply)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class Media(ElementBase):
    name = 'media'
    namespace = 'urn:xmpp:media-element'
    plugin_attrib = 'media'
    interfaces = {'height', 'width', 'alt'}

    def add_uri(self, value, itype):
        uri = URI()
        uri['value'] = value
        uri['type'] = itype
        self.append(uri)


class URI(ElementBase):
    name = 'uri'
    namespace = 'urn:xmpp:media-element'
    plugin_attrib = 'uri'
    plugin_multi_attrib = 'uris'
    interfaces = {'type', 'value'}

    def get_value(self):
        return self.xml.text

    def set_value(self, value):
        self.xml.text = value

    def del_value(self):
        self.xml.text = ''


register_stanza_plugin(Media, URI, iterable=True)



from slixmpp.plugins import BasePlugin

from . import stanza


class XEP_0402(BasePlugin):

    """
    XEP-0402: PEP Native bookmarks
    """

    name = "xep_0402"
    description = "XEP-0402: PEP Native bookmarks"
    dependencies = {"xep_0402"}
    stanza = stanza

    def plugin_init(self):
        stanza.register_plugin()



import unittest
from slixmpp.test.integration import SlixIntegration


class TestVcardTemp(SlixIntegration):
    async def asyncSetUp(self):
        await super().asyncSetUp()
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0054'])
        await self.connect_clients()

    async def _clear_vcard(self):
        # cleanup
        await self.clients[0]['xep_0054'].publish_vcard(
            self.clients[0]['xep_0054'].make_vcard()
        )

    async def test_vcard(self):
        """Check we can set and get a vcard"""
        await self._clear_vcard()

        # Check that vcard is empty
        recv = await self.clients[1]['xep_0054'].get_vcard(
            self.clients[0].boundjid.bare
        )
        self.assertEqual(recv['vcard_temp']['TITLE'], None)

        vcard = self.clients[0]['xep_0054'].make_vcard()
        vcard['TITLE'] = 'Coucou coucou'
        await self.clients[0]['xep_0054'].publish_vcard(
            vcard,
        )
        #
        recv = await self.clients[1]['xep_0054'].get_vcard(
            self.clients[0].boundjid.bare
        )
        self.assertEqual(recv['vcard_temp']['TITLE'], 'Coucou coucou')

        await self._clear_vcard()


suite = unittest.TestLoader().loadTestsFromTestCase(TestVcardTemp)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class GmailQuery(ElementBase):
    namespace = 'google:mail:notify'
    name = 'query'
    plugin_attrib = 'gmail'
    interfaces = set(['newer_than_time', 'newer_than_tid', 'search'])

    def get_search(self):
        return self._get_attr('q', '')

    def set_search(self, search):
        self._set_attr('q', search)

    def del_search(self):
        self._del_attr('q')

    def get_newer_than_time(self):
        return self._get_attr('newer-than-time', '')

    def set_newer_than_time(self, value):
        self._set_attr('newer-than-time', value)

    def del_newer_than_time(self):
        self._del_attr('newer-than-time')

    def get_newer_than_tid(self):
        return self._get_attr('newer-than-tid', '')

    def set_newer_than_tid(self, value):
        self._set_attr('newer-than-tid', value)

    def del_newer_than_tid(self):
        self._del_attr('newer-than-tid')


class MailBox(ElementBase):
    namespace = 'google:mail:notify'
    name = 'mailbox'
    plugin_attrib = 'gmail_messages'
    interfaces = set(['result_time', 'url', 'matched', 'estimate'])

    def get_matched(self):
        return self._get_attr('total-matched', '')

    def get_estimate(self):
        return self._get_attr('total-estimate', '') == '1'

    def get_result_time(self):
        return self._get_attr('result-time', '')


class MailThread(ElementBase):
    namespace = 'google:mail:notify'
    name = 'mail-thread-info'
    plugin_attrib = 'thread'
    plugin_multi_attrib = 'threads'
    interfaces = set(['tid', 'participation', 'messages', 'date',
                      'senders', 'url', 'labels', 'subject', 'snippet'])
    sub_interfaces = set(['labels', 'subject', 'snippet'])

    def get_senders(self):
        result = []
        senders = self.xml.findall('{%s}senders/{%s}sender' % (
            self.namespace, self.namespace))

        for sender in senders:
            result.append(MailSender(xml=sender))

        return result


class MailSender(ElementBase):
    namespace = 'google:mail:notify'
    name = 'sender'
    plugin_attrib = name
    interfaces = set(['address', 'name', 'originator', 'unread'])

    def get_originator(self):
        return self.xml.attrib.get('originator', '0') == '1'

    def get_unread(self):
        return self.xml.attrib.get('unread', '0') == '1'


class NewMail(ElementBase):
    namespace = 'google:mail:notify'
    name = 'new-mail'
    plugin_attrib = 'gmail_notification'


register_stanza_plugin(MailBox, MailThread, iterable=True)




# slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0353.stanza import Propose, Retract, Accept, Proceed, Reject
from slixmpp.plugins.xep_0353.jingle_message import XEP_0353

register_plugin(XEP_0353)



import unittest
from base64 import b64encode
from datetime import datetime
from pathlib import Path
from tempfile import NamedTemporaryFile
from hashlib import sha256

from slixmpp.plugins.xep_0082 import format_datetime
from slixmpp.test import SlixTest

class TestSIMS(SlixTest):
    def setUp(self):
        self.stream_start(
            mode="component", jid="whatevs.shakespeare.lit", plugins={"xep_0385"}
        )

    def test_set_file(self):
        with NamedTemporaryFile("wb+") as f:
            n = 10
            size = 0
            for i in range(n):
                size += len(bytes(i))
                f.write(bytes(i))

            f.seek(0)
            h = b64encode(sha256(f.read()).digest()).decode()
            sims = self.xmpp["xep_0385"].get_sims(
                Path(f.name),
                ["https://xxx.com"],
                media_type="MEDIA",
                desc="DESCRIPTION",
            )

            self.check(
                sims,
                f"""
                <reference xmlns='urn:xmpp:reference:0' type='data'>
                <media-sharing xmlns='urn:xmpp:sims:1'>
                  <file xmlns='urn:xmpp:jingle:apps:file-transfer:5'>
                    <media-type>MEDIA</media-type>
                    <name>{Path(f.name).name}</name>
                    <size>{size}</size>
                    <hash xmlns='urn:xmpp:hashes:2' algo='sha-256'>{h}</hash>
                    <desc>DESCRIPTION</desc>
                    <date>{format_datetime(datetime.fromtimestamp(Path(f.name).stat().st_mtime))}</date>
                  </file>
                  <sources>
                    <reference xmlns='urn:xmpp:reference:0' type='data' uri='https://xxx.com' />
                  </sources>
                </media-sharing>
                </reference>
                """,
                use_values=False,
            )


suite = unittest.TestLoader().loadTestsFromTestCase(TestSIMS)



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import sys
import logging
from getpass import getpass
from argparse import ArgumentParser

try:
    from httplib import HTTPSConnection
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode
    from http.client import HTTPSConnection

import asyncio
import slixmpp
from slixmpp.xmlstream import JID


class ThirdPartyAuthBot(slixmpp.ClientXMPP):

    """
    A simple Slixmpp bot that will echo messages it
    receives, along with a short thank you message.

    This version uses a thirdpary service for authentication,
    such as Facebook or Google.
    """

    def __init__(self, jid, password):
        slixmpp.ClientXMPP.__init__(self, jid, password)

        # The X-GOOGLE-TOKEN mech is ranked lower than PLAIN
        # due to Google only allowing a single SASL attempt per
        # connection. So PLAIN will be used for TLS connections,
        # and X-GOOGLE-TOKEN for non-TLS connections. To use
        # X-GOOGLE-TOKEN with a TLS connection, explicitly select
        # it using:
        #
        # slixmpp.ClientXMPP.__init__(self, jid, password,
        #                               sasl_mech="X-GOOGLE-TOKEN")

        # The session_start event will be triggered when
        # the bot establishes its connection with the server
        # and the XML streams are ready for use. We want to
        # listen for this event so that we we can initialize
        # our roster.
        self.add_event_handler("session_start", self.start)

        # The message event is triggered whenever a message
        # stanza is received. Be aware that that includes
        # MUC messages and error messages.
        self.add_event_handler("message", self.message)

    async def start(self, event):
        """
        Process the session_start event.

        Typical actions for the session_start event are
        requesting the roster and broadcasting an initial
        presence stanza.

        Arguments:
            event -- An empty dictionary. The session_start
                     event does not provide any additional
                     data.
        """
        self.send_presence()
        await self.get_roster()

    def message(self, msg):
        """
        Process incoming message stanzas. Be aware that this also
        includes MUC messages and error messages. It is usually
        a good idea to check the messages's type before processing
        or sending replies.

        Arguments:
            msg -- The received message stanza. See the documentation
                   for stanza objects and the Message stanza to see
                   how it may be used.
        """
        if msg['type'] in ('chat', 'normal'):
            msg.reply("Thanks for sending\n%(body)s" % msg).send()


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")


    access_token = None

    # Since documentation on how to work with Google tokens
    # can be difficult to find, we'll demo a basic version
    # here. Note that responses could refer to a Captcha
    # URL that would require a browser.

    # Using Facebook or MSN's custom authentication requires
    # a browser, but the process is the same once a token
    # has been retrieved.

    # Request an access token from Google:
    try:
        conn = HTTPSConnection('www.google.com')
    except:
        print('Could not connect to Google')
        sys.exit()

    params = urlencode({
        'accountType': 'GOOGLE',
        'service': 'mail',
        'Email': JID(args.jid).bare,
        'Passwd': args.password
    })
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded'
    }
    try:
        conn.request('POST', '/accounts/ClientLogin', params, headers)
        resp = conn.getresponse().read()
        data = {}
        for line in resp.split():
            k, v = line.split(b'=', 1)
            data[k] = v
    except Exception as e:
        print('Could not retrieve login data')
        sys.exit()

    if b'SID' not in data:
        print('Required data not found')
        sys.exit()


    params = urlencode({
        'SID': data[b'SID'],
        'LSID': data[b'LSID'],
        'service': 'mail'
    })
    try:
        conn.request('POST', '/accounts/IssueAuthToken', params, headers)
        resp = conn.getresponse()
        data = resp.read().split()
    except:
        print('Could not retrieve auth data')
        sys.exit()

    if not data:
        print('Could not retrieve token')
        sys.exit()

    access_token = data[0]


    # Setup the ThirdPartyAuthBot and register plugins. Note that while plugins
    # may have interdependencies, the order in which you register them does not
    # matter.

    # If using MSN, the JID should be "user@messenger.live.com", which will
    # be overridden on session bind.

    # We're using an access token instead of a password, so we'll use `''` as
    # a password argument filler.

    xmpp = ThirdPartyAuthBot(args.jid, '')
    xmpp.credentials['access_token'] = access_token

    # The credentials dictionary is used to provide additional authentication
    # information beyond just a password.

    xmpp.register_plugin('xep_0030') # Service Discovery
    xmpp.register_plugin('xep_0004') # Data Forms
    xmpp.register_plugin('xep_0060') # PubSub

    # MSN will kill connections that have been inactive for even
    # short periods of time. So use pings to keep the session alive;
    # whitespace keepalives do not work.
    xmpp.register_plugin('xep_0199', {'keepalive': True, 'frequency': 60})

    # If you are working with an OpenFire server, you may need
    # to adjust the SSL version used:
    # xmpp.ssl_version = ssl.PROTOCOL_SSLv3

    # If you want to verify the SSL certificates offered by a server:
    # xmpp.ca_certs = "path/to/ca/cert"

    # Connect to the XMPP server and start processing XMPP stanzas.
    # Google only allows one SASL attempt per connection, so in order to
    # enable the X-GOOGLE-TOKEN mechanism, we'll disable TLS.
    xmpp.connect()
    asyncio.get_event_loop().run_forever()



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza.rootstanza import RootStanza
from slixmpp.xmlstream import StanzaBase


class Presence(RootStanza):

    """
    XMPP's <presence> stanza allows entities to know the status of other
    clients and components. Since it is currently the only multi-cast
    stanza in XMPP, many extensions add more information to <presence>
    stanzas to broadcast to every entry in the roster, such as
    capabilities, music choices, or locations (XEP-0115: Entity Capabilities
    and XEP-0163: Personal Eventing Protocol).

    Since <presence> stanzas are broadcast when an XMPP entity changes
    its status, the bulk of the traffic in an XMPP network will be from
    <presence> stanzas. Therefore, do not include more information than
    necessary in a status message or within a <presence> stanza in order
    to help keep the network running smoothly.

    Example <presence> stanzas:

    .. code-block:: xml

        <presence />

        <presence from="user@example.com">
          <show>away</show>
          <status>Getting lunch.</status>
          <priority>5</priority>
        </presence>

        <presence type="unavailable" />

        <presence to="user@otherhost.com" type="subscribe" />

    Stanza Interface:
        - **priority**: A value used by servers to determine message routing.
        - **show**: The type of status, such as away or available for chat.
        - **status**: Custom, human readable status message.

    Attributes:
        - **types**: One of: available, unavailable, error, probe,
          subscribe, subscribed, unsubscribe, and unsubscribed.
        - **showtypes**: One of: away, chat, dnd, and xa.
    """

    name = 'presence'
    namespace = 'jabber:client'
    plugin_attrib = name
    interfaces = {'type', 'to', 'from', 'id', 'show', 'status', 'priority'}
    sub_interfaces = {'show', 'status', 'priority'}
    lang_interfaces = {'status'}

    types = {'available', 'unavailable', 'error', 'probe', 'subscribe',
             'subscribed', 'unsubscribe', 'unsubscribed'}
    showtypes = {'dnd', 'chat', 'xa', 'away'}

    def __init__(self, *args, recv: bool = False, **kwargs):
        """
        Initialize a new <presence /> stanza with an optional 'id' value.

        Overrides StanzaBase.__init__.
        """
        StanzaBase.__init__(self, *args, **kwargs)
        if not recv and self['id'] == '':
            if self.stream:
                use_ids = getattr(self.stream, 'use_presence_ids', None)
                if use_ids:
                    self['id'] = self.stream.new_id()

    def set_show(self, show: str):
        """
        Set the value of the <show> element.

        :param str show: Must be one of: away, chat, dnd, or xa.
        """
        if show is None:
            self._del_sub('show')
        elif show in self.showtypes:
            self._set_sub_text('show', text=show)
        return self

    def get_type(self) -> str:
        """
        Return the value of the <presence> stanza's type attribute, or
        the value of the <show> element if valid.
        """
        out = self._get_attr('type')
        if not out and self['show'] in self.showtypes:
            out = self['show']
        if not out or out is None:
            out = 'available'
        return out

    def set_type(self, value: str):
        """
        Set the type attribute's value, and the <show> element
        if applicable.

        :param str value: Must be in either self.types or self.showtypes.
        """
        if value in self.types:
            self['show'] = None
            if value == 'available':
                value = ''
            self._set_attr('type', value)
        elif value in self.showtypes:
            self['show'] = value
        return self

    def del_type(self):
        """
        Remove both the type attribute and the <show> element.
        """
        self._del_attr('type')
        self._del_sub('show')

    def set_priority(self, value: int):
        """
        Set the entity's priority value. Some server use priority to
        determine message routing behavior.

        Bot clients should typically use a priority of 0 if the same
        JID is used elsewhere by a human-interacting client.

        :param int value: An integer value greater than or equal to 0.
        """
        self._set_sub_text('priority', text=str(value))

    def get_priority(self):
        """
        Return the value of the <presence> element as an integer.

        :rtype: int
        """
        p = self._get_sub_text('priority')
        if not p:
            p = 0
        try:
            return int(p)
        except ValueError:
            # The priority is not a number: we consider it 0 as a default
            return 0

    def reply(self, clear=True):
        """
        Create a new reply <presence/> stanza from ``self``.

        Overrides StanzaBase.reply.

        :param bool clear: Indicates if the stanza contents should be removed
                           before replying. Defaults to True.
        """
        new_presence = StanzaBase.reply(self, clear)
        if self['type'] == 'unsubscribe':
            new_presence['type'] = 'unsubscribed'
        elif self['type'] == 'subscribe':
            new_presence['type'] = 'subscribed'
        return new_presence



"""
This only covers the component registration side of the XEP-0077 plugin
"""

import unittest

from slixmpp import ComponentXMPP, Iq
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0077 import Register


class TestRegistration(SlixTest):
    def setUp(self):
        self.stream_start(
            mode="component", plugins=["xep_0077"], jid="shakespeare.lit", server="lit"
        )

    def testRegistrationForm(self):
        self.stream_start(
            mode="component", plugins=["xep_0077"], jid="shakespeare.lit", server="lit"
        )
        self.recv(
            """
            <iq type='get' id='reg1' to='shakespeare.lit' from='bill@server/resource'>
                <query xmlns='jabber:iq:register'/>
            </iq>
            """,
        )
        self.send(
            f"""
            <iq type='result' id='reg1' from='shakespeare.lit' to='bill@server/resource'>
                <query xmlns='jabber:iq:register'>
                    <instructions>{self.xmpp["xep_0077"].form_instructions}</instructions>
                    <username/>
                    <password/>
                </query>
            </iq>
            """,
            use_values=False  # Fails inconsistently without this
        )

    def testRegistrationSuccessAndModif(self):
        self.recv(
            """
            <iq type='set' id='reg2' to='shakespeare.lit' from="bill@server/resource">
                <query xmlns='jabber:iq:register'>
                    <username>bill</username>
                    <password>Calliope</password>
                </query>
            </iq>
            """
        )
        self.send("<iq type='result' id='reg2' from='shakespeare.lit' to='bill@server/resource'/>")
        user_store = self.xmpp["xep_0077"]._user_store
        self.assertEqual(user_store["bill@server"]["username"], "bill")
        self.assertEqual(user_store["bill@server"]["password"], "Calliope")

        self.recv(
            """
            <iq type='get' id='reg1' to='shakespeare.lit' from="bill@server/resource">
                <query xmlns='jabber:iq:register'/>
            </iq>
            """,
        )
        self.send(
            f"""
            <iq type='result' id='reg1' to="bill@server/resource" from='shakespeare.lit'>
                <query xmlns='jabber:iq:register'>
                    <instructions>{self.xmpp["xep_0077"].form_instructions}</instructions>
                    <username>bill</username>
                    <password>Calliope</password>
                    <registered />
                </query>
            </iq>
            """,
            use_values=False  # Fails inconsistently without this
        )

    def testRegistrationAndRemove(self):
        self.recv(
            """
            <iq type='set' id='reg2' to='shakespeare.lit' from="bill@shakespeare.lit/globe">
                <query xmlns='jabber:iq:register'>
                    <username>bill</username>
                    <password>Calliope</password>
                </query>
            </iq>
            """
        )
        self.send("<iq type='result' id='reg2' from='shakespeare.lit' to='bill@shakespeare.lit/globe'/>")
        pseudo_iq = self.xmpp.Iq()
        pseudo_iq["from"] = "bill@shakespeare.lit/globe"
        fut = self.xmpp.wrap(self.xmpp["xep_0077"].api["user_get"](None, None, None, pseudo_iq))
        self.run_coro(fut)
        user = fut.result()
        self.assertEqual(user["username"], "bill")
        self.assertEqual(user["password"], "Calliope")
        self.recv(
            """
            <iq type='set' from='bill@shakespeare.lit/globe' id='unreg1'>
            <query xmlns='jabber:iq:register'>
                <remove/>
            </query>
            </iq>
            """
        )
        self.send("<iq type='result' to='bill@shakespeare.lit/globe' id='unreg1'/>")
        user_store = self.xmpp["xep_0077"]._user_store
        self.assertIs(user_store.get("bill@shakespeare.lit"), None)



suite = unittest.TestLoader().loadTestsFromTestCase(TestRegistration)
